import collections.abc

import numpy as np
from pycce.bath.array import check_gyro, point_dipole, BathArray, _add_args, _stevens_str_doc
from pycce.bath.map import InteractionMap
from pycce.constants import ELECTRON_GYRO
from pycce.h.total import central_hamiltonian
from pycce.sm import _smc
from pycce.utilities import expand, rotate_coordinates, rotate_tensor, outer, \
    normalize


def _attr_arr_setter(self, attr, value, dtype=np.float64):
    if getattr(self, attr) is not None:
        obj = getattr(self, attr)
        obj[...] = np.asarray(value, dtype=dtype)
    else:
        setattr(self, attr, np.asarray(value, dtype=dtype))


class Center:
    r"""
    Class, which contains the properties of the single central spin. Should *not* be initialized directly - use
    ``CenterArray`` instead.

    Args:
        position (ndarray with shape (3, )):
            Cartesian coordinates in Angstrom of the central spin. Default (0., 0., 0.).

        spin (float): Total spin of the central spin.

        D (float or ndarray with shape (3, )): D (longitudinal splitting) parameter of central spin
            in ZFS tensor of central spin in kHz.

            *OR*

            Total ZFS tensor. Default 0.

        E (float): E (transverse splitting) parameter of central spin in ZFS tensor of central spin in kHz.
            Default 0. Ignored if ``D`` is None or tensor.

        gyro (float or ndarray with shape (3, 3))): Gyromagnetic ratio of central spin in rad / ms / G.

            *OR*

            Tensor describing central spin interactions with the magnetic field.

            Default -17608.597050 kHz * rad / G - gyromagnetic ratio of the free electron spin.

        alpha (int or ndarray with shape (2*spin + 1, )):
            :math:`\ket{0}` state of the qubit in :math:`S_z` basis or the index of eigenstate to be used as one.

            Default is **None**.

        beta (int or ndarray with shape (2*spin + 1, )):
            :math:`\ket{1}` state of the qubit in :math:`S_z` basis or the index of eigenstate to be used as one.

            Default is **None**.

        detuning (float): Energy detuning from the zeeman splitting in kHz,
            included as an extra :math:`+\omega \hat S_z` term in the Hamiltonian,
            where :math:`\omega` is the detuning.

            Default 0.
    """
    _state_mapping = {0: 'beta', 1: 'alpha'}

    def __init__(self, position=None,
                 spin=0, D=0, E=0,
                 gyro=ELECTRON_GYRO, alpha=None, beta=None, detuning=0):

        if position is None:
            position = np.array([0, 0, 0])

        self._zfs = None
        self._gyro = None

        self._xyz = None
        self._s = None
        self._h = {}

        self._detuning = None

        self.xyz = position
        self.s = spin
        self.set_zfs(D, E)
        self.set_gyro(gyro)
        self.detuning = detuning

        self.projections_alpha = None
        r"""ndarray with shape (3,): Vector with spin operator matrix elements
        of type :math:`[\bra{0}\hat S_x\ket{0}, \bra{0}\hat S_y\ket{0}, \bra{0}\hat S_z\ket{0}]`, where
        :math:`\ket{0}` is the alpha qubit state. Generated by ``CenterArray``."""
        self.projections_beta = None
        r"""ndarray with shape (3,): Vector with spin operator matrix elements
        of type :math:`[\bra{1}\hat S_x\ket{1}, \bra{1}\hat S_y\ket{1}, \bra{1}\hat S_z\ket{1}]`, where
        :math:`\ket{1}` is the beta qubit state. Generated by ``CenterArray``."""
        # You cannot initialize these from center, only from CenterArray
        self.projections_alpha_all = None
        r"""
        ndarray with shape (2s-1, 3):
            Array of vectors of the central spin matrix elements of form:

            .. math::

                [\bra{0}\hat{S}_x\ket{j}, \bra{0}\hat{S}_y\ket{j}, \bra{0}\hat{S}_z\ket{j}],

            where :math:`\ket{0}` is the alpha qubit state, and :math:`\ket{\j}` are all states.
        """

        self.projections_beta_all = None
        r"""
        ndarray with shape (2s-1, 3):
            Array of vectors of the central spin matrix elements of form:

            .. math::

                [\bra{1}\hat{S}_x\ket{j}, \bra{1}\hat{S}_y\ket{j}, \bra{1}\hat{S}_z\ket{j}],

            where :math:`\ket{1}` is the beta qubit state, and :math:`\ket{\j}` are all states.
        """

        self.energies = None
        """ndarray with shape (2s-1,): Array of the central spin Hamiltonian eigen energies."""
        self.eigenvectors = None
        """ndarray: Eigen states of the central spin Hamiltonian."""

        self.hamiltonian = None
        """Hamiltonian: Central spin Hamiltonian."""

        self._alpha = None
        self._beta = None

        self.alpha_index = None
        """int: Index of the central spin Hamiltonian eigen state, chosen as alpha state of the qubit."""
        self.beta_index = None
        """int: Index of the central spin Hamiltonian eigen state, chosen as beta state of the qubit."""

        self.alpha = alpha
        self.beta = beta

        self._sigma = None

    def get_projections(self, which):
        if which == 'alpha' or which == 1:
            return self.projections_alpha
        elif which == 'beta' or which == 0:
            return self.projections_beta

    def get_projections_all(self, which):
        if which == 'alpha' or which == 1:
            return self.projections_alpha_all
        elif which == 'beta' or which == 0:
            return self.projections_beta_all

    @property
    def xyz(self):
        """ndarray with shape (3, ): Position of the central spin in Cartesian coordinates."""
        return self._xyz

    @xyz.setter
    def xyz(self, position):
        _attr_arr_setter(self, '_xyz', position)

    @property
    def gyro(self):
        """
        ndarray with shape (3,3 ) or (n, 3, 3): Tensor describing central spin interactions
            with the magnetic field or array of spins.

            Default -17608.597050 rad / ms / G - gyromagnetic ratio of the free electron spin."""
        gyro, _ = check_gyro(self._gyro)
        return gyro

    @gyro.setter
    def gyro(self, gyro):
        _attr_arr_setter(self, '_gyro', gyro)

    @property
    def zfs(self):
        """ndarray with shape (3, 3) or (n, 3, 3): Zero field splitting tensor of the central spin or array of spins."""
        return self._zfs

    @zfs.setter
    def zfs(self, zfs):
        _attr_arr_setter(self, '_zfs', zfs)

    @property
    def s(self):
        """float or ndarray with shape (n, ): Total spin of the central spin or array of spins."""
        return self._s[()]

    @s.setter
    def s(self, spin):
        _attr_arr_setter(self, '_s', spin)

    @property
    def detuning(self):
        """ndarray with shape (3, ): Position of the central spin in Cartesian coordinates."""
        return self._detuning[()]

    @_add_args(_stevens_str_doc)
    @property
    def h(self):
        return self._h

    @detuning.setter
    def detuning(self, detune):
        _attr_arr_setter(self, '_detuning', detune)

    def set_zfs(self, D=0, E=0):
        """
         Set Zero Field Splitting of the central spin from longitudinal ZFS *D* and transverse ZFS *E*.

        Args:
            D (float or ndarray with shape (3, 3)): D (longitudinal splitting) parameter of central spin
                in ZFS tensor of central spin in kHz.

                **OR**

                Total ZFS tensor. Default 0.

            E (float): E (transverse splitting) parameter of central spin in ZFS tensor of central spin in kHz.
                 Default 0. Ignored if ``D`` is None or tensor.
        """

        self.zfs = zfs_tensor(D, E)

    def set_gyro(self, gyro):
        """
        Set gyromagnetic ratio of the central spin.

        Args:
            gyro (float or ndarray with shape (3, 3)): Gyromagnetic ratio of central spin in rad / ms / G.

                **OR**

                Tensor describing central spin interactions with the magnetic field.

                Default -17608.597050 kHz * rad / G - gyromagnetic ratio of the free electron spin.

        """
        check = not np.asarray(gyro).shape == (3, 3)
        if check:
            gyro, check = check_gyro(gyro)

            if check:
                gyro = np.eye(3) * gyro

        self._gyro = gyro

    @property
    def alpha(self):
        r"""
        ndarray or int: :math:`\ket{0}` qubit state of the central spin in :math:`S_z` basis

        **OR**

        index of the energy state to be considered as one.
        """

        return self._get_state('alpha')

    @alpha.setter
    def alpha(self, state):
        self._set_state('alpha', state)

    @property
    def beta(self):
        r"""
        ndarray or int: :math:`\ket{1}` qubit state of the central spin in :math:`S_z` basis

        **OR**

        index of the energy state to be considered as one.
        """
        return self._get_state('beta')

    @beta.setter
    def beta(self, state):
        self._set_state('beta', state)

    @property
    def dim(self):
        """int or ndarray with shape (n, ): Dimensions of the central spin or array of spins."""

        return (self._s * 2 + 1 + 1e-8).astype(int)[()]

    def generate_sigma(self):
        r"""
        Generate Pauli matrices of the qubit in :math:`S_z` basis.
        """
        assert np.isclose(np.inner(self.alpha.conj(), self.beta), 0), \
            f"Pauli matrix can be generated only for orthogonal states, " \
            f"{self.alpha} and {self.beta} are not orthogonal"

        alpha_x_alpha = outer(self.alpha, self.alpha)
        beta_x_beta = outer(self.beta, self.beta)
        alpha_x_beta = outer(self.alpha, self.beta)
        beta_x_alpha = outer(self.beta, self.alpha)

        self._sigma = {'x': alpha_x_beta + beta_x_alpha,
                       'y': -1j * alpha_x_beta + 1j * beta_x_alpha,
                       'z': alpha_x_alpha - beta_x_beta}

    @property
    def sigma(self):
        """
        dict: Dictionary with Pauli matrices of the qubit in :math:`S_z` basis.
        """
        if self._sigma is None:
            self.generate_sigma()
        return self._sigma

    def _set_state(self, name, state):
        if state is not None:
            state = np.asarray(state)
            if state.size == 1 and np.issubdtype(state.dtype, int):
                setattr(self, name + '_index', int(state))
            elif state.dtype == bool:
                setattr(self, name + '_index', state)
            else:
                assert state.size == np.prod(self.dim), f"Incorrect format of {name}: {state}"
                setattr(self, '_' + name, normalize(state))

                # remove index if manually set alpha state
                setattr(self, name + '_index', None)
        else:
            setattr(self, '_' + name, state)
        self._sigma = None

    def _get_state(self, name):

        state = getattr(self, '_' + name)
        if state is not None:
            return state

        state = getattr(self, name + '_index')
        if state is not None:
            return state

        return None

    def get_state(self, name):
        if not isinstance(name, str):
            name = self._state_mapping[name]
        return self._get_state(name)

    def generate_states(self, magnetic_field=None, bath=None, projected_bath_state=None):
        r"""
        Compute eigenstates of the central spin Hamiltonian.

        Args:
            magnetic_field (ndarray with shape (3,)): Array containing external magnetic field as (Bx, By, Bz).
            bath (BathArray with shape (m,) or ndarray with shape (m, 3, 3):
                Array of all bath spins or array of hyperfine tensors.
            projected_bath_state (ndarray with shape (m,) or (m, 3)):
                Array of :math:`I_z` projections for each bath spin.
        """

        self.generate_hamiltonian(magnetic_field=magnetic_field, bath=bath, projected_bath_state=projected_bath_state)
        self.energies, self.eigenvectors = np.linalg.eigh(self.hamiltonian, UPLO='U')

        if self.alpha_index is not None:
            self._alpha = np.ascontiguousarray(self.eigenvectors[:, self.alpha_index])
        if self.beta_index is not None:
            self._beta = np.ascontiguousarray(self.eigenvectors[:, self.beta_index])

    def __repr__(self):
        message = f"{self.__class__.__name__}" + ("\n(s: " + self.s.__str__() +
                                                  ",\nxyz:\n" + self.xyz.__str__() +
                                                  ",\nzfs:\n" + self.zfs.__str__() +
                                                  ",\ngyro:\n" + self.gyro.__str__())
        if self._detuning.any():
            message += "\ndetuning: " + self.gyro.__str__()
        message += ")"
        return message

    def generate_hamiltonian(self, magnetic_field=None, bath=None, projected_bath_state=None):
        r"""
        Generate central spin Hamiltonian.

        Args:
            magnetic_field (ndarray with shape (3, ) or func):
                Magnetic field of type ``magnetic_field = np.array([Bx, By, Bz])``
                or callable with signature ``magnetic_field(pos)``, where ``pos`` is ndarray with shape (3, ) with the
                position of the spin.

            bath (BathArray with shape (n,) or ndarray with shape (n, 3, 3)):
                Array of bath spins or hyperfine tensors.

            projected_bath_state (ndarray with shape (n, )): :math:`S_z` projections of the bath spin states.

        Returns:
            Hamiltonian: Central spin Hamiltonian, including
                first order contributions from the bath spins.
        """
        if magnetic_field is None:
            magnetic_field = np.array([0., 0., 0.], dtype=np.float64)

        if not callable(magnetic_field):
            magnetic_field = np.asarray(magnetic_field)
            if magnetic_field.size == 1:
                magnetic_field = np.array([0., 0., magnetic_field.reshape(-1)[0]])

        if isinstance(bath, BathArray):
            bath = bath.A
            projected_bath_state = bath.proj

        self.hamiltonian = central_hamiltonian(self, magnetic_field, hyperfine=bath,
                                               bath_state=projected_bath_state)
        return self.hamiltonian

    def transform(self, rotation=None, style='col'):
        """
        Apply coordinate transformation to the central spin.

        Args:
            rotation (ndarray with shape (3, 3)): Rotation matrix.
            style (str): Can be 'row' or 'col'. Determines how rotation matrix is initialized.

        """
        self.xyz = rotate_coordinates(self.xyz, rotation=rotation, style=style)
        self.zfs = rotate_tensor(self.zfs, rotation=rotation, style=style)
        self._gyro = rotate_tensor(self._gyro, rotation=rotation, style=style)
        return


class CenterArray(Center, collections.abc.Sequence):
    r"""
    Class, containing properties of all central spins. The properties of the each separate spin can be accessed
    as elements of the object directly. Each element of the array is an instance of the ``Center`` class.

    Examples:

        Generate array of 2 electron central spins:

        >>> import numpy as np
        >>> ca = CenterArray(2, spin=0.5) # Array of size 2 with spins-1/2
        >>> print(ca)
        CenterArray
        (s: [0.5 0.5],
        xyz:
        [[0. 0. 0.]
         [0. 0. 0.]],
        zfs:
        [[[0. 0. 0.]
          [0. 0. 0.]
          [0. 0. 0.]]
         [[0. 0. 0.]
          [0. 0. 0.]
          [0. 0. 0.]]],
        gyro:
        [[[-17608.59705     -0.          -0.     ]
          [    -0.      -17608.59705     -0.     ]
          [    -0.          -0.      -17608.59705]]
         [[-17608.59705     -0.          -0.     ]
          [    -0.      -17608.59705     -0.     ]
          [    -0.          -0.      -17608.59705]]])

        Set first two eigenstates of the combined central spin Hamiltonian as a singlie qubit state:

        >>> ca.alpha = 0
        >>> ca.beta = 1

        Change gyromagnetic ratio of the first spin:

        >>> ca[0].gyro = np.eye(3) * 1000
        >>> print(ca[0])
        Center
        (s: 0.5,
        xyz:
        [0. 0. 0.],
        zfs:
        [[0. 0. 0.]
         [0. 0. 0.]
         [0. 0. 0.]],
        gyro:
        1000.0)


    Args:
        size (int): Number of central spins.

        spin (ndarray with shape (size,)):
            Total spins of the central spins.

            .. note::

                All center spin properties are broadcasted to the total size of the center array,
                provided by ``size`` argument, or inferred from ``spin``, ``position`` arguments.

        position (ndarray with shape (size, 3)):
            Cartesian coordinates in Angstrom of the central spins. Default (0., 0., 0.).

        D (ndarray with shape (size, ) or ndarray with shape (n, 3, 3)):
            D (longitudinal splitting) parameters of central spins
            in ZFS tensor of central spin in kHz.

            *OR*

            Total ZFS tensor. Default 0.

        E (ndarray with shape (size, )):
            E (transverse splitting) parameters of central spins
            in ZFS tensor of central spin in kHz.
            Default 0. Ignored if ``D`` is None or tensor.

        gyro (ndarray with shape (size, ) or ndarray with shape (size, 3, 3))):
            Gyromagnetic ratios of the central spins in rad / ms / G.

            *OR*

            Tensors describing central spins interactions with the magnetic field.

            Default -17608.597050 kHz * rad / G - gyromagnetic ratio of the free electron spin.

        imap (dict or InteractionMap or ndarray with shape (3, 3)):
            Dict-like object containing interaction tensors between the central spins of the structure {(i, j): T_ij}.
            Where i, j are positional indexes of the central spins. If provided as an ndarray with shape (3, 3),
            assumes the same interactions between all pairs of central spins in the array.
            If provided with shape (size * (size - 1) / 2, 3, 3), assigns the interactions to the ordered pairs:
            ``{(0, 1): imap[0], (0, 2): imap[1] ... (size - 2, size - 1): imap[-1]}``


        alpha (int or ndarray with shape (S, )):
            :math:`\ket{0}` state of the qubit in the product space of all central spins,
            or the index of eigenstate to be used as one.

            Default is **None**.

        beta (int or ndarray with shape (S, )):
            :math:`\ket{1}` state of the qubit in the product space of all central spins,
            or the index of eigenstate to be used as one.

            Default is **None**.

        detuning (ndarray with shape (size, )): Energy detunings from the Zeeman splitting in kHz,
            included as an extra :math:`+\omega \hat S_z` term in the Hamiltonian,
            where :math:`\omega` is the detuning.

            Default is 0.
    """

    def __init__(self, size=None, position=None,
                 spin=None, D=0, E=0,
                 gyro=ELECTRON_GYRO, imap=None,
                 alpha=None,
                 beta=None,
                 detuning=0):

        if size is None:
            if spin is not None:
                spin = np.asarray(spin)
                size = spin.size

            elif position is not None:
                position = np.asarray(position)
                size = position.size // 3

            else:
                raise ValueError('Size of the array is not provided')

        self.size = size

        if position is None:
            position = np.asarray([[0, 0, 0]] * self.size)
        if spin is None:
            spin = 0

        position = np.asarray(position)

        spin = np.asarray(spin).reshape(-1)

        if spin.size != self.size:
            spin = np.array(np.broadcast_to(spin, self.size))

        if position.ndim == 1:
            position = np.array(np.broadcast_to(position, (self.size, position.size)))

        detuning = np.asarray(detuning).reshape(-1)

        if detuning.size != self.size:
            detuning = np.array(np.broadcast_to(detuning, self.size))

        self._state = None
        self.state_index = None

        super().__init__(position=position, spin=spin, D=D, E=E, gyro=gyro, alpha=alpha, beta=beta, detuning=detuning)

        self._array = np.array([Center(position=p, spin=s[..., 0], D=zfs, gyro=g, detuning=d) for p, s, zfs, g, d in
                                zip(self.xyz, self.s[:, np.newaxis], self.zfs, self.gyro, self.detuning)],
                               dtype=object)

        self._h = np.asarray([x.h for x in self._array], dtype=object)

        if isinstance(imap, dict):
            imap = InteractionMap.from_dict(imap)

        if imap is not None and not isinstance(imap, InteractionMap):

            if self.size < 2:
                raise ValueError(f'Cannot assign interaction map for array of size {self.size}')

            imap = np.broadcast_to(imap, (self.size * (self.size - 1) // 2, 3, 3))
            imap = InteractionMap(rows=np.arange(self.size), columns=np.arange(1, self.size), tensors=imap)

        self._imap = imap

        self.energy_alpha = None
        """float: Energy of the alpha state. Generated by ``.generate_projections`` call if ``second_order=True``."""
        self.energy_beta = None
        """float: Energy of the beta state. Generated by ``.generate_projections`` call if ``second_order=True``."""

        self.energies = None
        """ndarray with shape (n, ): Energy of each eingenstate of the central spin Hamiltonian."""

    @property
    def imap(self):
        """
        InteractionMap: dict-like object, which contains interactions between central spins.
        """
        if self._imap is None:
            self._imap = InteractionMap()

        return self._imap

    @property
    def alpha(self):
        r"""
        ndarray or int: :math:`\ket{0}` qubit state of the central spin in :math:`S_z` basis

        **OR** index of the energy state to be considered as one.

        If not provided in the ``CentralArray`` instance, returns the tensor product of all ``alpha`` states of
        each element of the array. If there are undefined ``alpha`` states of the elements of the array,
        raises an error.

        Examples:

            >>> ca = CenterArray(2, spin=0.5) # Array of size 2 with spins-1/2
            >>> ca[0].alpha = [0,1]
            >>> ca[1].alpha = [1,0]
            >>> print(ca.alpha)
            [0.+0.j 0.+0.j 1.+0.j 0.+0.j]

        """

        return self._get_state('alpha')

    @alpha.setter
    def alpha(self, state):
        self._set_state('alpha', state)

    @property
    def beta(self):
        r"""
        ndarray or int: :math:`\ket{1}` qubit state of the central spin in :math:`S_z` basis

        **OR** index of the energy state to be considered as one.
        """
        return self._get_state('beta')

    @beta.setter
    def beta(self, state):
        self._set_state('beta', state)

    @property
    def state(self):
        r"""
        ndarray: Initial state of the qubit in gCCE simulations.
        Assumed to be :math:`\frac{1}{\sqrt{2}}(\ket{0} + \ket{1})` unless provided."""
        state = super(CenterArray, self)._get_state('state')

        if isinstance(state, np.ndarray) and state.dtype == bool:
            state = state.reshape(-1, 2)

            alpha = 1
            beta = 1

            for c, s in zip(self, state):
                alpha = np.kron(alpha, c.get_state(s[0]))
                beta = np.kron(beta, c.get_state(s[1]))

            return normalize(alpha + beta)

        if state is not None:
            return state

        else:
            self._check_states()
            return normalize(self.alpha + self.beta)

    @state.setter
    def state(self, state):
        self._set_state('state', state)

    @property
    def gyro(self):
        return self._gyro

    @gyro.setter
    def gyro(self, gyro):
        _attr_arr_setter(self, '_gyro', gyro)

    def __getitem__(self, item):
        newarray = self._array.__getitem__(item)
        if isinstance(newarray, Center):
            return newarray

        else:
            xyz = self.xyz[item]
            gyro = self.gyro[item]
            s = self.s[item]
            zfs = self.zfs[item]
            ca = CenterArray(len(newarray), position=xyz, gyro=gyro, spin=s, D=zfs)
            ca._array = newarray

            if self._imap is not None:
                ca._imap = self.imap.subspace(np.arange(self.size)[item])

            if self.h.any():
                ca._h = self.h[item]

            return ca

    def __setitem__(self, key, val):
        if not isinstance(val, Center):
            raise ValueError

        self.zfs[key] = val.zfs
        self.gyro[key] = val.gyro
        self.s[key] = val.s
        self.xyz[key] = val.xyz

        center = self._array.__getitem__(key)

        center.alpha = val.alpha
        center.beta = val.beta
        center.h.update(val.h)

    def __len__(self):
        return self.size

    def set_zfs(self, D=0, E=0):

        darr = np.asarray(D)

        if darr.shape == (self.size, 3, 3):
            self.zfs = darr

        else:
            if self.zfs is None:
                self.zfs = np.zeros((self.size, 3, 3), dtype=np.float64)

            for i, (d, e) in enumerate(zip(np.broadcast_to(D, self.size), np.broadcast_to(E, self.size))):
                self.zfs[i] = zfs_tensor(d, e)

    def set_gyro(self, gyro):

        garr = np.asarray(gyro)

        if garr.shape == (self.size, 3, 3):
            self.gyro = garr
        else:
            if self.gyro is None:
                self.gyro = np.zeros((self.size, 3, 3), dtype=np.float64)

            for i, g in enumerate(np.broadcast_to(gyro, self.size)):

                g, check = check_gyro(g)

                if check:
                    self._gyro[i] = np.eye(3) * g
                else:
                    self._gyro[i] = g

    def point_dipole(self):
        """
        Using point-dipole approximation, generate interaction tensors between central spins.
        """
        for i in range(self.size):
            for j in range(i + 1, self.size):
                c1 = self[i]
                c2 = self[j]
                self.imap[i, j] = point_dipole(c1.xyz - c2.xyz, c1.gyro, c2.gyro)

    def generate_states(self, magnetic_field=None, bath=None, projected_bath_state=None):
        if isinstance(bath, BathArray):
            projected_bath_state = bath.proj
            bath = bath.A

        for i, c in enumerate(self):
            if bath is None:
                hf = None
            elif len(self) == 1:
                hf = bath
            else:
                hf = bath[..., i, :, :]

            c.generate_states(magnetic_field=magnetic_field,
                              bath=hf, projected_bath_state=projected_bath_state)

        super(CenterArray, self).generate_states(magnetic_field=magnetic_field,
                                                 bath=bath, projected_bath_state=projected_bath_state)
        if self.state_index is not None and isinstance(self.state_index, (int, np.integer)):
            self._state = np.ascontiguousarray(self.eigenvectors[:, self.state_index])

    def generate_projections(self, second_order=False, level_confidence=0.95):
        r"""
        Generate vectors with the spin projections of the spin states:

            .. math::

                [\bra{a}\hat{S}_x\ket{a}, \bra{a}\hat{S}_y\ket{a}, \bra{a}\hat{S}_z\ket{a}],

        where :math:`\ket{a}` and  is alpha or beta qubit state.
        They are stored in the ``.projections_alpha`` and ``.projections_beta`` respectively.

        If ``second_order`` is set to ``True``, also generates matrix elements of qubit states and all
        other eigenstates of the central spin Hamiltonian, used in computing second order couplings between bath spins:

        .. math::

                [\bra{a}\hat{S}_x\ket{b}, \bra{a}\hat{S}_y\ket{b}, \bra{a}\hat{S}_z\ket{b}],

        where :math:`\ket{a}` is qubit level and :math:`\ket{b}` are all other energy levels.

        This function is called in the ``CCE`` routine.

        .. note::

            if qubit state are not eigenstates and ``second_order`` set to ``True``, for each qubit state finds a close
            eigenstate (with minimal fidelity between two states set by ``level_confidence`` keyword) and uses that
            one instead of user provided.


        Args:
            second_order (bool): True if generate properties, necessary for second order corrections.
            level_confidence (float):
                Minimum fidelity between an eigenstate and provided qubit level for them to be
                considered the same. Used only if ``second_order == True``.

        """
        self._check_states()

        if second_order:

            ai = _close_state_index(self.alpha, self.eigenvectors, level_confidence=level_confidence)
            bi = _close_state_index(self.beta, self.eigenvectors, level_confidence=level_confidence)

            alpha = self.eigenvectors[:, ai]
            beta = self.eigenvectors[:, bi]

            self.energy_alpha = self.energies[ai]
            self.energy_beta = self.energies[bi]

            self.energies = self.energies

            gp = generate_projections
            self.projections_alpha_all = np.array([gp(alpha, s, spins=self.s) for s in self.eigenvectors.T])
            self.projections_beta_all = np.array([gp(beta, s, spins=self.s) for s in self.eigenvectors.T])

            for i, center in enumerate(self):
                center.projections_alpha_all = self.projections_alpha_all[:, i]
                center.projections_beta_all = self.projections_beta_all[:, i]

        else:

            self.energy_alpha = None
            self.energy_beta = None

            self.projections_alpha_all = None
            self.projections_beta_all = None

        self.projections_alpha = np.array(generate_projections(self.alpha, spins=self.s))
        self.projections_beta = np.array(generate_projections(self.beta, spins=self.s))

        for i, center in enumerate(self):
            center.projections_alpha = self.projections_alpha[i]
            center.projections_beta = self.projections_beta[i]

    def get_energy(self, which):
        r"""
        Get energy of the qubit state.

        Args:
            which (str): ``alpha`` for :math:`\ket{0}` qubit state, ``beta`` for :math:`\ket{1}` qubit state.

        Returns:
            float: Energy of the qubit state.
        """

        if which == 'alpha' or which == True:
            return self.energy_alpha
        elif which == 'beta' or which == False:
            return self.energy_beta

    def generate_sigma(self):
        self._check_states()
        super(CenterArray, self).generate_sigma()

        for i, c in enumerate(self):
            if c.alpha is not None and c.beta is not None:
                try:
                    c.generate_sigma()
                    for x in c.sigma:
                        c.sigma[x] = expand(c.sigma[x], i, self.dim)
                except AssertionError:
                    pass

    def _check_states(self):
        for n in ['alpha', 'beta']:
            s = getattr(self, n)
            if s is None or isinstance(s, int):
                raise ValueError(f'Wrong {n} format: {s}')

    def _get_state(self, name):
        state = getattr(self, '_' + name)
        if state is not None:
            return state

        state = getattr(self, name + '_index')
        if state is not None:
            return state

        state = 1
        for c in self:
            s = getattr(c, name)
            if s is None:
                return None
            state = np.kron(state, s)
        state = normalize(state)
        return state

    def add_interaction(self, i, j, tensor):
        """
        Add interactions tensor between bath spins with indexes ``i`` and ``j``.

        Args:
            i (int or ndarray (n,) ):
                Index of the first spin in the pair or array of the indexes of the first spins in n pairs.
            j (int or ndarray with shape (n,)):
                Index of the second spin in the pair or array of the indexes of the second spins in n pairs.
            tensor (ndarray with shape (3,3) or (n, 3,3)):
                Interaction tensor between the spins i and j or array of tensors.
        """

        self.imap[i, j] = tensor


def _close_state_index(state, eiv, level_confidence=0.95):
    r"""
    Get index of the eigenstate stored in eiv,
    which has fidelity higher than ``level_confidence`` with the provided ``state``.

    Args:
        state (ndarray with shape (2s+1,)): State for which to find the analogous eigen state.
        eiv (ndarray with shape (2s+1, 2s+1)): Matrix of eigenvectors as columns.
        level_confidence (float): Threshold fidelity. Default 0.95.

    Returns:
        int: Index of the eigenstate.
    """
    ev_state = eiv.conj().T @ state

    indexes = np.argwhere(ev_state * ev_state.conj() > level_confidence).flatten()

    if not indexes.size:
        raise ValueError(f"Initial qubit state is below F = {level_confidence} "
                         f"to the eigenstate of central spin Hamiltonian.\n"
                         f"Qubit level:\n{repr(state)}\n"
                         f"Eigenstates (rows):\n{repr(eiv.T)}")
    return indexes[0]


def generate_projections(state_a, state_b=None, spins=None):
    r"""
    Generate vector or list of vectors (if ``spins`` is not None) with the spin projections of the given spin states:

    .. math::

        [\bra{a}\hat{S}_x\ket{b}, \bra{a}\hat{S}_y\ket{b}, \bra{a}\hat{S}_z\ket{b}],

    where :math:`\ket{a}` and :math:`\ket{b}` are the given spin states.

    Args:
        state_a (ndarray): State :math:`\ket{a}` of the central spin or spins in :math:`\hat{S}_z` basis.
        state_b (ndarray): State :math:`\ket{b}` of the central spin or spins in :math:`\hat{S}_z` basis.
        spins (ndarray, optional): Array of spins, comprising the given state vectors.
            If provided, assumes that states correspond to a Hilbert space of several spins, and projections
            of states are computed for each spin separately.
    Returns:
        ndarray with shape (3,) or list:
            :math:`[\braket{\hat{S}_x}, \braket{\hat{S}_y}, \braket{\hat{S}_z}]` projections or list of projections.
    """
    if state_b is None:
        state_b = state_a
    if spins is None:
        spin = (state_a.size - 1) / 2
        sm = _smc[spin]

        projections = np.array([state_a.conj() @ sm.x @ state_b,
                                state_a.conj() @ sm.y @ state_b,
                                state_a.conj() @ sm.z @ state_b],
                               dtype=np.complex128)
    else:
        projections = []
        dim = (np.asarray(spins) * 2 + 1 + 1e-8).astype(int)

        for i, s in enumerate(spins):
            sm = _smc[s]
            smx = expand(sm.x, i, dim)
            smy = expand(sm.y, i, dim)
            smz = expand(sm.z, i, dim)

            p = np.array([state_a.conj() @ smx @ state_b,
                          state_a.conj() @ smy @ state_b,
                          state_a.conj() @ smz @ state_b],
                         dtype=np.complex128)
            projections.append(p)

    return projections


def zfs_tensor(D, E=0):
    """
    Generate (3, 3) ZFS tensor from observable parameters D and E.

    Args:
        D (float or ndarray with shape (3, 3)): Longitudinal splitting (D) in ZFS **OR** total ZFS tensor.
        E (float): Transverse splitting (E) in ZFS.

    Returns:
        ndarray with shape (3, 3): Total ZFS tensor.
    """
    D = np.asarray(D)

    if D.size == 1:

        tensor = np.zeros((3, 3), dtype=np.float64)
        tensor[2, 2] = 2 / 3 * D
        tensor[1, 1] = -D / 3 - E
        tensor[0, 0] = -D / 3 + E

    else:
        tensor = D

    return tensor



import numpy as np

PI2 = 2 * np.pi

# Energy
MHZ_TO_KHZ = 1000

joule_to_khz = 1.50919E+30
hartree_to_khz = 6.57968E+12
rydberg_to_khz = 3.28984E+12
ev_to_khz = 2.41799E+11
kjpermole_to_khz = 2.50607E+09
kcalpermole_to_khz = 1.04854E+10
hz_to_khz = 1.00000E-03
khz_to_khz = 1.
mhz_to_khz = 1.00000E+03
ghz_to_khz = 1.00000E+06
thz_to_khz = 1.00000E+09
phz_to_khz = 1.00000E+12
wavenumber_to_khz = 2.99792E+07

_CONVERSION_TO_RADKHZ = {
    'ghz': ghz_to_khz * PI2,
    'radghz': ghz_to_khz,
    'mhz': mhz_to_khz * PI2,
    'radmhz': mhz_to_khz,
    'khz': khz_to_khz * PI2,
    'radkhz': khz_to_khz,
    'hz': hz_to_khz * PI2,
    'radhz': hz_to_khz,
    'joule': joule_to_khz * PI2,
}

m_to_angstrom = 1.00000E+10
bohr_to_angstrom = 5.29177E-01
a_to_angstrom = 1

_CONVERSION_TO_A = {
    'bohr': bohr_to_angstrom,
    'm': m_to_angstrom,
    'nm': m_to_angstrom * 1e-9,
    'a': a_to_angstrom
}

BOHR_TO_ANGSTROM = 5.29177E-01

HARTREE_TO_MHZ = 6579680000.0
M_TO_BOHR = 18897300000.0

ELECTRON_GYRO = -17608.597050  # rad / (ms * Gauss) or rad * kHz / G
HBAR_MU0_O4PI = 1.05457172  # When everything else in rad, kHz, ms, G, A

COMPLEX_DTYPE = np.complex128

BARN_TO_BOHR2 = M_TO_BOHR ** 2 * 1E-28
EFG_CONVERSION = BARN_TO_BOHR2 * HARTREE_TO_MHZ * MHZ_TO_KHZ  # units to convert EFG

PI2 = np.pi * 2
HBAR_SI = 6.62607015e-34 / PI2
BOHR_MAGNETON = 9.274009994E-24
NUCLEAR_MAGNETON = 5.05078366E-27



"""
Module with helper functions to obtain CPMG coherence from the noise autocorrelation function.
"""
import numba
import numpy as np
import scipy.integrate
from numba import cfunc, carray
from numba.types import intc, CPointer, float64
from scipy import LowLevelCallable

from .constants import PI2


def _jit_integrand_function(integrand_function):
    jitted_function = numba.jit(integrand_function, nopython=True)

    @cfunc(float64(intc, CPointer(float64)))
    def wrapped(n, xx):
        values = carray(xx, n)
        return jitted_function(values)

    return LowLevelCallable(wrapped.ctypes)


@cfunc(float64(float64, float64, float64))
def _yfunc(x, tau, npulses):
    delay = tau / 2 / npulses
    pulses_passed = (x // delay + 1) // 2

    if pulses_passed % 2 == 0:
        return 1
    else:
        return -1


@_jit_integrand_function
def _integrand(args):
    v, u, tau, np = args[0], args[1], args[2], args[3]
    return _yfunc((v + u) / 2, tau, np) * _yfunc((v - u) / 2, tau, np)


def filterfunc(ts, tau, npulses):
    """
    Time-domain filter function for the given CPMG sequence.

    Args:
        ts (ndarray with shape (n,)): Time points at which filter function will be computed.
        tau (float): Delay between pulses.
        npulses (int): Number of pulses in CPMG sequence.

    Returns:
        ndarray with shape (n,): Filter function for the given CPMG sequence.
    """
    fs = np.empty(ts.shape)
    ks = np.r_[np.arange(npulses), np.arange(npulses)]

    for i, u in enumerate(ts):
        bad_points = (2 * ks + 1) * tau / npulses
        bad_points[:npulses] += u
        bad_points[npulses:] -= u
        bad_points = np.unique(bad_points[(bad_points > u) & (bad_points < 2 * tau - u)])

        if bad_points.size == 0:
            fs[i] = scipy.integrate.quad(_integrand, u, 2 * tau - u, args=(u, tau, npulses))[0]

        else:
            fs[i] = scipy.integrate.quad(_integrand, u, bad_points[0], args=(u, tau, npulses))[0]

            for j in range(1, bad_points.size):
                fs[i] += scipy.integrate.quad(_integrand, bad_points[j - 1], bad_points[j],
                                              args=(u, tau, npulses))[0]

            fs[i] += scipy.integrate.quad(_integrand, bad_points[-1], 2 * tau - u, args=(u, tau, npulses))[0]

    return fs


def gaussian_phase(timespace, corr, npulses, units='khz'):
    """
    Compute average random phase squared assuming Gaussian noise.

    Args:
        timespace (ndarray with shape (n,)): Time points at which correlation function was computed.
        corr (ndarray with shape (n,)): Noise autocorrelation function.
        npulses (int): Number of pulses in CPMG sequence.
        units (str): If units contain frequency or angular frequency ('rad' in ``units``).
    Returns:
        ndarray with shape (n,): Random phase accumulated by the qubit.
    """
    if 'rad' not in units:
        corr *= PI2 ** 2
    timespace = np.asarray(timespace)
    chis = np.zeros(timespace.shape, dtype=np.complex128)
    for i, tau in enumerate(timespace):
        if tau == 0:
            chis[i] = 0
        else:
            chis[i] = np.trapz(corr[timespace <= tau] * filterfunc(timespace[timespace <= tau], tau, npulses),
                               timespace[timespace <= tau]) / 2
    return chis



from collections.abc import MutableMapping

import numpy as np
from numba import jit
from pycce.utilities import expand


class SpinMatrix:
    """
    Class containing the spin matrices in Sz basis.

    Args:
        s (float): Total spin.
    """

    def __init__(self, s):
        dim = int(2 * s + 1 + 1e-8)

        projections = np.linspace(-s, s, dim, dtype=np.complex128)

        self.plus = np.zeros((dim, dim), dtype=np.complex128)
        self.p = self.plus

        for i in range(dim - 1):
            self.plus[i, i + 1] += np.sqrt(s * (s + 1) -
                                           projections[i] * projections[i + 1])

        self.minus = self.plus.conj().T
        self.m = self.minus

        self.s = s
        self.dim = dim

        self.x = 1 / 2. * (self.plus + self.minus)
        self.y = 1 / 2j * (self.plus - self.minus)
        self.z = np.diag(projections[::-1])

        self.eye = np.eye(dim, dtype=np.complex128)

        self._stevens = {}

    def stev(self, k, q):
        try:
            return self._stevens[(k, q)]
        except KeyError:
            self._stevens[(k, q)] = stevo(self, k, q)

        return self._stevens[(k, q)]

    def __repr__(self):
        return "Spin-{:.1f} matrices x, y, z".format(self.s)


class MatrixDict(MutableMapping):
    """
    Class for storing the SpinMatrix objects.

    """

    def __init__(self, *spins):
        self._data = {}
        if spins:
            for s in spins:
                self[s] = SpinMatrix[s]

    def __getitem__(self, key):
        try:
            cond = key in self._data.keys()
        except TypeError:
            key = key[()]
            cond = key in self._data.keys()
        if not cond:
            self._data[key] = SpinMatrix(key)
        return self._data[key]

    def __delitem__(self, key):
        del self._data[key]

    def __setitem__(self, key, value):
        if not isinstance(value, SpinMatrix):
            raise ValueError('Illegal definition of MatrixDict element. Only accepts SpinMatrix objects')
        self._data[key] = value

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return f"{type(self).__name__}(" + " ".join(str(x) for x in self.keys()) + ")"

    def keys(self):
        return self._data.keys()


_smc = MatrixDict()

# Stevens operators
from scipy.special import comb


def _a(k, q, m, j=1., i=1):
    if k == q and m == 0:
        return 1
    elif q + m > k or m < 0:
        return 0

    q = q + 1

    first = (2 * q + m - 1) * _a(k, q, m - 1, j=j, i=i) + (q * (q - 1) - m * (m + 1) / 2) * _a(k, q, m, j=j, i=i)
    second = 0

    for n in range(1, k - q - m + 1):
        stuff = comb(m + n, m) * (j * (j + 1)) ** i - comb(m + n, m - 1) - comb(m + n, m - 2)
        second += (-1) ** n * stuff * _a(k, q, m + n, j=j, i=i)

    return first + second


def _f(k, q):
    a_s = []
    for m in range(0, k - q + 1):
        for i in range(0, int((k - q - m) / 2 + 1)):
            a_s.append(_a(k, q, m, j=1, i=i))
    a_s = np.abs(np.array(a_s, dtype=int))
    a_s = a_s[a_s > 0]

    return np.gcd.reduce(a_s)


def stevo(sm, k, q):
    """
    Stevens operators (from I.D. Ryabov, Journal of Magnetic Resonance 140, 141–145 (1999)).

    Args:
        sm (SpinMatrix): Spin matrices of the given spin.
        k (int): :math:`k` index of the Stevens operator.
        q (int): :math:`q` index of the Stevens operator.

    Returns:
        ndarray with shape (n, n): Stevens operator representation in :math:`S_z` basis.
    """
    spin = sm.s
    mp = np.linalg.matrix_power
    if q == 0:
        alpha = 2
    elif (k % 2 == 1) or (q % 2 == 0):
        alpha = 1
    else:
        alpha = 1 / 2

    sign = np.sign(q)
    if q < 0:
        q = -q
        pref = alpha / (2j * _f(k, q))

    else:
        pref = alpha / (2 * _f(k, q))

    full = 0
    for m in range(0, k - q + 1):
        full += _a(k, q, m, j=spin) * ((mp(sm.plus, q) + sign * (-1) ** (k - q - m) * mp(sm.minus, q)) @ mp(sm.z, m))

    return pref * full


def dimensions_spinvectors(bath=None, central_spin=None):
    """
    Generate two arrays, containing dimensions of the spins in the cluster and the vectors with spin matrices.

    Args:
        bath (BathArray with shape (n,)): Array of the n spins within cluster.
        central_spin (CenterArray, optional): If provided, include dimensions of the central spins.

    Returns:
        tuple: *tuple* containing:

            * **ndarray with shape (n,)**: Array with dimensions for each spin.

            * **list**: List with vectors of spin matrices for each spin in the cluster
              (Including central spin if ``central_spin`` is not None). Each with  shape (3, N, N) where
              ``N = prod(dimensions)``.
    """
    dimensions = []

    if bath is not None:
        dimensions += [n.dim for n in bath]

    if central_spin is not None:
        try:
            for c in central_spin:
                dimensions += [c.dim]

        except TypeError:
            dimensions += [central_spin.dim]

    dimensions = np.array(dimensions, dtype=np.int32)
    vectors = vecs_from_dims(dimensions)

    return dimensions, vectors


@jit(cache=True, nopython=True)
def vecs_from_dims(dimensions):
    """
    Generate ndarray of spin vectors, given the array of spin dimensions.

    Args:
        dimensions (ndarray with shape (n,)): Dimensions of spins.

    Returns:
        ndarray with shape (n, 3, X, X): Array of spin vectors in full Hilbert space.
    """
    td = dimensions.prod()
    vectors = np.zeros((len(dimensions), 3, td, td), dtype=np.complex128)
    for j, d in enumerate(dimensions):
        vectors[j] = spinvec(j, dimensions)
    return vectors


@jit(cache=True, nopython=True)
def spinvec(j, dimensions):
    """
    Generate single spin vector, given the index and dimensions of all spins in the cluster.

    Args:
        j (int): Index of the spin.
        dimensions (ndarray with shape (n,)): Dimensions of spins.

    Returns:
        ndarray with shape (3, X, X): Spin vector of :math:`j`-sth spin in full Hilbert space.

    """
    x, y, z = numba_gen_sm(dimensions[j])
    vec = np.stack((expand(x, j, dimensions),
                    expand(y, j, dimensions),
                    expand(z, j, dimensions))
                   )
    return vec


@jit(cache=True, nopython=True)
def numba_gen_sm(dim):
    """
    Numba-friendly spin matrix.
    Args:
        dim (int): dimensions of the spin marix.

    Returns:
        ndarray:
    """
    s = (dim - 1) / 2
    projections = np.linspace(-s, s, dim).astype(np.complex128)
    plus = np.zeros((dim, dim), dtype=np.complex128)

    for i in range(dim - 1):
        plus[i, i + 1] += np.sqrt(s * (s + 1) -
                                  projections[i] * projections[i + 1])

    minus = plus.conj().T
    x = 1 / 2. * (plus + minus)
    y = 1 / 2j * (plus - minus)
    z = np.diag(projections[::-1])
    return x, y, z



import numpy as np
from numba import jit, generated_jit, types

from pycce.constants import HBAR_MU0_O4PI, PI2, ELECTRON_GYRO
from pycce.utilities import tensor_vdot, vec_tensor_vec


@jit(cache=True, nopython=True)
def expanded_single(ivec, gyro, mfield, self_tensor, detuning=.0):
    """
    Function to compute the single bath spin term.

    Args:
        ivec (ndarray with shape (3, n, n)): Spin vector of the bath spin in the full Hilbert space of the cluster.
        gyro (float or ndarray with shape (3, 3)):
        mfield (ndarray wtih shape (3, ): Magnetic field of type ``mfield = np.array([Bx, By, Bz])``.
        self_tensor (ndarray with shape (3, 3)): tensor of self-interaction of type IPI where I is bath spin.
        detuning (float): Additional term of d*Iz allowing to simulate different energy splittings of bath spins.

    Returns:
        ndarray with shape (n, n): Single bath spin term.

    """
    hzeeman = zeeman(ivec, gyro, mfield)
    hself = np.zeros(hzeeman.shape, dtype=hzeeman.dtype)

    if ivec[2, 0, 0].real > 0.5:
        hself = vec_tensor_vec(ivec, self_tensor, ivec)

        # v_ivec = np.einsum('ij,jkl->ikl', self_tensor, ivec, dtype=np.complex128)
        # hself = np.einsum('lij,ljk->ik', ivec, v_ivec, dtype=np.complex128)

    if detuning:
        hself += detuning * ivec[2]

    return hself + hzeeman


@generated_jit(cache=True, nopython=True)
def zeeman(ivec, gyro, mfield):
    """
    Function
    Args:
        ivec (ndarray with shape (3, n, n)): Spin vector of the  spin in the full Hilbert space of the cluster.
        gyro (float or ndarray with shape (3, 3)): Gyromagnetic ratio of the spin.
        mfield (ndarray with shape (3, )): Magnetic field at the position of the spin.

    Returns:
        ndarray with shape (n, n): Zeeman interactions.

    """
    if isinstance(gyro, types.Float):
        def z(ivec, gyro, mfield):
            return - gyro / PI2 * (mfield[0] * ivec[0] + mfield[1] * ivec[1] + mfield[2] * ivec[2])

    # else assume tensor
    else:
        def z(ivec, gyro, mfield):
            gsvec = tensor_vdot(gyro / PI2, ivec)
            return mfield[0] * gsvec[0] + mfield[1] * gsvec[1] + mfield[2] * gsvec[2]

    return z


@generated_jit(cache=True, nopython=True)
def dd_tensor(coord_1, coord_2, g1, g2):
    """
    Generate dipole-dipole interaction tensor.

    Args:
        coord_1 (ndarray with shape (3,)): Coordinates of the first spin.
        coord_2 (ndarray with shape (3,)): Coordinates of the second spin.
        g1 (float or ndarray with shape (3, 3)): Gyromagnetic ratio of the first spin.
        g2 (float or ndarray with shape (3, 3)): Gyromagnetic ratio of the second spin.

    Returns:
        ndarray with shape (3, 3): Interaction tensor.

    """
    if isinstance(g2, types.Float) and isinstance(g1, types.Float):
        def func(coord_1, coord_2, g1, g2):
            p_tensor = gen_pos_tensor(coord_1, coord_2) * HBAR_MU0_O4PI / PI2
            p_tensor *= g2 * g1
            return p_tensor

    elif isinstance(g1, types.Float):
        def func(coord_1, coord_2, g1, g2):
            p_tensor = gen_pos_tensor(coord_1, coord_2) * HBAR_MU0_O4PI / PI2
            p_tensor = g1 * p_tensor @ g2
            return p_tensor
    elif isinstance(g2, types.Float):
        def func(coord_1, coord_2, g1, g2):
            p_tensor = gen_pos_tensor(coord_1, coord_2) * HBAR_MU0_O4PI / PI2
            p_tensor = g1 @ p_tensor * g2
            return p_tensor
    else:
        def func(coord_1, coord_2, g1, g2):
            p_tensor = gen_pos_tensor(coord_1, coord_2) * HBAR_MU0_O4PI / PI2
            p_tensor = g1 @ p_tensor @ g2
            return p_tensor
    return func


@jit(cache=True, nopython=True)
def dipole_dipole(coord_1, coord_2, g1, g2, ivec_1, ivec_2):
    """
    Compute dipole_dipole interactions between two bath spins.

    Args:
        coord_1 (ndarray with shape (3,)): Coordinates of the first spin.
        coord_2 (ndarray with shape (3,)): Coordinates of the second spin.
        g1 (float): Gyromagnetic ratio of the first spin.
        g2 (float): Gyromagnetic ratio of the second spin.
        ivec_1 (ndarray with shape (3, n, n)): Spin vector of the first spin in the full Hilbert space of the cluster.
        ivec_2 (ndarray with shape (3, n, n)): Spin vector of the second spin in the full Hilbert space of the cluster.

    Returns:
        ndarray with shape (n, n): Dipole-dipole interactions.

    """
    p_tensor = dd_tensor(coord_1, coord_2, g1, g2)
    return vec_tensor_vec(ivec_1, p_tensor, ivec_2)


@jit(cache=True, nopython=True)
def gen_pos_tensor(coord_1, coord_2):
    """
    Generate positional tensor -(3*r @ r.T - r*r), used for hyperfine tensor (without gyro factor).

    Args:
        coord_1 (ndarray with shape (3,)): Coordinates of the first spin.
        coord_2 (ndarray with shape (3,)): Coordinates of the second spin.

    Returns:
        ndarray with shape (3, 3): Positional tensor.

    """
    pos = coord_1 - coord_2
    r = np.linalg.norm(pos)

    return -(3 * np.outer(pos, pos) - np.eye(3, dtype=np.complex128) * r ** 2) / (r ** 5)


def bath_interactions(nspin, ivectors):
    """
    Compute interactions between bath spins.

    Args:
        nspin (BathArray): Array of the bath spins in the given cluster.
        ivectors (array-like): array of expanded spin vectors, each with shape (3, n, n).

    Returns:
        ndarray with shape (n, n): All intrabath interactions of bath spins in the cluster.

    """

    nnuclei = len(nspin)
    imap = nspin.imap
    dd = 0
    if imap is None:
        for i in range(nnuclei):
            for j in range(i + 1, nnuclei):
                n1 = nspin[i]
                n2 = nspin[j]

                ivec_1 = ivectors[i]
                ivec_2 = ivectors[j]
                dd += dipole_dipole(n1['xyz'], n2['xyz'], n1.gyro, n2.gyro, ivec_1, ivec_2)
    else:
        for i in range(nnuclei):
            for j in range(i + 1, nnuclei):
                n1 = nspin[i]
                n2 = nspin[j]

                ivec_1 = ivectors[i]
                ivec_2 = ivectors[j]

                try:
                    tensor = imap[i, j]
                    dd += vec_tensor_vec(ivec_1, tensor, ivec_2)
                except KeyError:
                    dd += dipole_dipole(n1['xyz'], n2['xyz'], n1.gyro, n2.gyro, ivec_1, ivec_2)
    return dd


@jit(cache=True, nopython=True)
def bath_mediated(hyperfines, ivectors, energy_state, energies, projections):
    r"""
    Compute all hyperfine-mediated interactions between bath spins.

    Args:
        hyperfines (ndarray with shape (n, 3, 3)): Array of hyperfine tensors of the bath spins in the given cluster.
        ivectors (array-like): array of expanded spin vectors, each with shape (3, n, n).
        energy_state (float): Energy of the qubit state on which the interaction is conditioned.
        energies (ndarray with shape (2s - 1,)): Array of energies of all states of the central spin.
        projections (ndarray with shape (2s - 1, 3)):
            Array of vectors of the central spin matrix elements of form:

            .. math::

                [\bra{i}\hat{S}_x\ket{j}, \bra{i}\hat{S}_y\ket{j}, \bra{i}\hat{S}_z\ket{j}],

            where  :math:`\ket{i}` are different states of the central spin.

    Returns:
        ndarray with shape (n, n): Hyperfine-mediated interactions.

    """
    mediated = np.zeros(ivectors[0, 0].shape, dtype=np.complex128)

    others_mask = energies != energy_state
    energies = energies[others_mask]
    projections = projections[others_mask]

    for energy_j, s_ij in zip(energies, projections):
        element_ij = np.zeros(ivectors[0, 0].shape, dtype=np.complex128)
        element_ji = np.zeros(ivectors[0, 0].shape, dtype=np.complex128)

        for hf, ivec in zip(hyperfines, ivectors):
            element_ij += conditional_hyperfine(hf, ivec, s_ij)
            element_ji += conditional_hyperfine(hf, ivec, s_ij.conj())
        mediated += element_ij @ element_ji / (energy_state - energy_j)

    return mediated


@jit(cache=True, nopython=True)
def conditional_hyperfine(hyperfine_tensor, ivec, projections):
    r"""
    Compute conditional hyperfine Hamiltonian.

    Args:
        hyperfine_tensor (ndarray with shape (3, 3)): Tensor of hyperfine interactions of the bath spin.
        ivec (ndarray with shape (3, n, n)): Spin vector of the bath spin in the full Hilbert space of the cluster.
        projections (ndarray with shape (3,)):
            Array of vectors of the central spin matrix elements of form:

            .. math::

                [\bra{i}\hat{S}_x\ket{j}, \bra{i}\hat{S}_y\ket{j}, \bra{i}\hat{S}_z\ket{j}],

            where :math:`\ket{j}` are different states of the central spin.
            If :math:`\ket{i} = \ket{j}`, produces the usual conditioned hyperfine interactions
            and just equal to projections of :math:`\hat{S}_z` of the central spin state
            :math:`[\braket{\hat{S}_x}, \braket{\hat{S}_y}, \braket{\hat{S}_z}]`.

            If :math:`\ket{i} \neq \ket{j}`, gives second order perturbation.

    Returns:
        ndarray with shape (n, n): Conditional hyperfine interaction.

    """

    a_ivec = tensor_vdot(hyperfine_tensor, ivec)
    return projections[0] * a_ivec[0] + projections[1] * a_ivec[1] + projections[2] * a_ivec[2]


@jit(cache=True, nopython=True)
def hyperfine(hyperfine_tensor, svec, ivec):
    """
    Compute hyperfine interactions between central spin and bath spin.

    Args:
        hyperfine_tensor (ndarray with shape (3, 3)): Tensor of hyperfine interactions of the bath spin.
        svec (ndarray with shape (3, n, n)): Spin vector of the central spin in the full Hilbert space of the cluster.
        ivec (ndarray with shape (3, n, n)): Spin vector of the bath spin in the full Hilbert space of the cluster.

    Returns:
        ndarray with shape (n, n): Hyperfine interaction.

    """
    return vec_tensor_vec(svec, hyperfine_tensor, ivec)


@jit(cache=True, nopython=True)
def self_central(svec, mfield, tensor, gyro=ELECTRON_GYRO, detuning=0):
    r"""
    Function to compute the central spin term in the Hamiltonian.

    Args:
        svec (ndarray with shape (3, n, n)): Spin vector of the central spin in the full Hilbert space of the cluster.
        mfield (ndarray wtih shape (3,): Magnetic field of type ``mfield = np.array([Bx, By, Bz])``.
        tensor (ndarray with shape (3, 3)):
            Zero Field Splitting tensor of the central spin.
        gyro (float or ndarray with shape (3,3)):
            gyromagnetic ratio of the central spin OR tensor corresponding to interaction between magnetic field and
            central spin.
        detuning (float): Energy detuning from the Zeeman splitting in kHz.

    Returns:
        ndarray with shape (n, n): Central spin term.

    """
    hzeeman = zeeman(svec, gyro, mfield)
    hself = np.zeros(hzeeman.shape, dtype=hzeeman.dtype)

    if svec[2, 0, 0].real > 0.5:
        hself = vec_tensor_vec(svec, tensor, svec)

    if detuning:
        hself += detuning * svec[2]

    return hself + hzeeman


def center_interactions(center, vectors):
    r"""
    Compute interactions between central spins.

    Args:
        center (CenterArray): Array of central spins
        vectors (ndarray with shape (x, 3, n, n)): Array of spin vectors of central spins.

    Returns:
        ndarray with shape (n, n): Central spin Overhauser term.

    """
    ncenters = len(center)
    if (ncenters == 1) or (not center.imap):
        return 0

    ham = 0

    for i in range(ncenters):
        for j in range(i + 1, ncenters):
            try:
                vec_1 = vectors[i]
                vec_2 = vectors[j]
                tensor = center.imap[i, j]
                ham += vec_tensor_vec(vec_1, tensor, vec_2)
            except KeyError:
                pass
    return ham


@jit(cache=True, nopython=True)
def overhauser_central(svec, others_hyperfines, others_state):
    """
    Compute Overhauser field term on the central spin from all other spins, not included in the cluster.

    Args:
        svec (ndarray with shape (3, n, n)): Spin vector of the central spin in the full Hilbert space of the cluster.
        others_hyperfines (ndarray with shape (m, 3, 3)):
            Array of hyperfine tensors for all bath spins not included in the cluster.
        others_state (ndarray with shape (m,) or (m, 3)):
            Array of :math:`I_z` projections for each bath spin outside of the given cluster.

    Returns:
        ndarray with shape (n, n): Central spin Overhauser term.

    """

    zfield = (others_hyperfines[..., 2, 2] * others_state).sum()

    return zfield * svec[2]


@jit(cache=True, nopython=True)
def overhauser_bath(ivec, position, gyro,
                    other_gyros, others_position, others_state):
    """
    Compute Overhauser field term on the bath spin in the cluster from all other spins, not included in the cluster.

    Args:
        ivec (ndarray with shape (3, n, n)): Spin vector of the bath spin in the full Hilbert space of the cluster.
        position (ndarray with shape (3,)): Position of the bath spin.
        gyro (float): Gyromagnetic ratio of the bath spin.
        other_gyros (ndarray with shape (m,)):
            Array of the gyromagnetic ratios of the bath spins, not included in the cluster.
        others_position (ndarray with shape (m, 3)):
            Array of the positions of the bath spins, not included in the cluster.
        others_state (ndarray with shape (m,) or (m, 3)):
            Array of :math:`I_z` projections for each bath spin outside of the given cluster.

    Returns:
        ndarray with shape (n, n): Bath spin Overhauser term.

    """

    pre = np.asarray(gyro * other_gyros * HBAR_MU0_O4PI / PI2)

    pos = position - others_position
    r = np.sqrt((pos ** 2).sum(axis=-1))
    zfield = (pre / r ** 5 * (r ** 2 - 3 * pos[..., 2] ** 2) * others_state).sum()

    return zfield * ivec[2]  # + xfield * ivec[0] + yfield * ivec[1]


@jit(cache=True, nopython=True)
def overhauser_from_tensors(vec, tensors, projected_state):
    """
    Compute Overhauser field from array of tensors.

    Args:
        vec (ndarray with shape (3, n, n)): Spin vector of the bath spin in the full Hilbert space of the cluster.

        tensors (ndarray with shape (N, 3, 3)): Array of interaction tensors.

        projected_state (ndarray with shape (N, )):
            Array of :math:`I_z` projections of the spins outside of the given cluster..

    Returns:
        ndarray with shape (n, n): Bath spin Overhauser term.

    """
    return (tensors[:, 2, 2] * projected_state).sum() * vec[2]


def projected_addition(vectors, bath, center, state):
    r"""
    Compute the first order addition of the interactions with the cental spin to the cluster Hamiltonian.

    Args:
        vectors (array-like): Array of expanded spin vectors, each with shape (3, n, n).
        bath (BathArray): Array of bath spins.
        center (CenterArray): Array of central spins.
        state (str, bool, or array-like): Identificator of the qubit spin. ``'alpha'`` or ``True``
            for :math:`\ket{0}` state, ``'beta'`` of ``False`` for :math:`\ket{1}` state.

    Returns:
        ndarray with shape (n, n): Addition to the Hamiltonian.
    """
    ncenters = len(center)
    addition = 0

    for ivec, n in zip(vectors, bath):
        try:

            iterator = iter(state)  # iterable

        except TypeError:
            for i, c in enumerate(center):
                if ncenters > 1:
                    hf = n.A[i]
                else:
                    hf = n.A
                projections = c.get_projections(state)
                addition += conditional_hyperfine(hf, ivec, projections)
        else:

            for i, (c, s) in enumerate(zip(center, iterator)):
                if ncenters > 1:
                    hf = n.A[i]
                else:
                    hf = n.A

                projections = c.get_projections(s)
                addition += conditional_hyperfine(hf, ivec, projections)

    energy = center.get_energy(state)

    if energy is not None:
        for i, c in enumerate(center):
            if ncenters > 1:
                hf = bath.A[:, i]
            else:
                hf = bath.A

            projections_state_all = c.get_projections_all(state)

            addition += bath_mediated(hf, vectors, energy,
                                      center.energies, projections_state_all)

    return addition


def center_external_addition(vectors, cluster, outer_spin, outer_state):
    """
    Compute the first order addition of the interactions between central spin and external bath spins
    to the cluster Hamiltonian.

    Args:
        vectors (array-like): Array of expanded spin vectors, each with shape (3, n, n).
        cluster (BathArray): Array of cluster spins.
        outer_spin (BathArray with shape (o, )): Array of the spins outside the cluster.
        outer_state (ndarray with shape (o, )): Array of the :math:`S_z` projections of the external bath spins.

    Returns:
        ndarray with shape (n, n): Addition to the Hamiltonian.
    """
    addition = 0
    ncenters = vectors.shape[0] - cluster.size

    for i, v in enumerate(vectors[cluster.size:]):
        if ncenters == 1:
            hf = outer_spin.A
        else:
            hf = outer_spin.A[:, i]

        addition += overhauser_central(v, hf, outer_state)

    return addition


def bath_external_point_dipole(vectors, cluster, outer_spin, outer_state):
    """
    Compute the first order addition of the point-dipole interactions between cluster spins and external bath spins
    to the cluster Hamiltonian.

    Args:
        vectors (array-like): Array of expanded spin vectors, each with shape (3, n, n).
        cluster (BathArray): Array of cluster spins.
        outer_spin (BathArray with shape (o, )): Array of the spins outside the cluster.
        outer_state (ndarray with shape (o, )): Array of the :math:`S_z` projections of the external bath spins.

    Returns:
        ndarray with shape (n, n): Addition to the Hamiltonian.
    """
    addition = 0
    for ivec, n in zip(vectors, cluster):
        addition += overhauser_bath(ivec, n.xyz, n.gyro, outer_spin.gyro,
                                    outer_spin.xyz, outer_state)
    return addition


def external_spins_field(vectors, indexes, bath, projected_state):
    """
    Compute the first order addition of the point-dipole interactions between cluster spins and external bath spins
    to the cluster Hamiltonian.

    Args:
        vectors (array-like): Array of expanded spin vectors, each with shape (3, n, n).
        indexes (ndarray with shape (n,)): Array of indexes of bath spins inside the given cluster.
        bath (BathArray with shape (N,)): Array of all bath spins.
        projected_state (ndarray with shape (N, ):  Array of the :math:`S_z` projections of all bath spins.

    Returns:
        ndarray with shape (n, n): Addition to the Hamiltonian.
    """
    outer_mask = np.ones(bath.size, dtype=bool)
    outer_mask[indexes] = False

    outer_spin = bath[outer_mask]
    outer_state = projected_state[outer_mask]
    cluster = bath[indexes]

    addition = center_external_addition(vectors, cluster, outer_spin, outer_state)

    if bath.imap is None:
        addition += bath_external_point_dipole(vectors, cluster, outer_spin, outer_state)

        return addition

    imap_indexes = bath.imap.indexes
    remove_j = np.ones(indexes.shape, dtype=bool)

    for j, (ind, ivec) in enumerate(zip(indexes, vectors)):

        outer_mask[:] = True
        outer_mask[indexes] = False

        where_index = (imap_indexes == ind)

        if where_index.any():
            remove_j[j] = False
            which_pairs = where_index.any(axis=1) & (~np.isin(imap_indexes, indexes[remove_j]).any(axis=1))
            remove_j[j] = True

            other_indexes = imap_indexes[which_pairs][~(where_index[which_pairs])]

            addition += overhauser_from_tensors(ivec, bath.imap.data[which_pairs], projected_state[other_indexes])

            outer_mask[other_indexes] = False

        if outer_mask.any():
            outer_spin = bath[outer_mask]
            outer_state = projected_state[outer_mask]
            n = bath[ind]
            addition += overhauser_bath(ivec, n.xyz, n.gyro, outer_spin.gyro,
                                        outer_spin.xyz, outer_state)

    return addition



import numpy as np
from numba import jit
from numba.typed import List
from pycce.sm import _smc
from pycce.utilities import gen_state_list, vector_from_s


class BathState:
    r"""
    Class for storing the state of the bath spins. Usually is not generated directly, but accessed as
    an ``BathArray.state`` attribute.

    Args:
        size (int): Number of bath states to be stored.
    """
    _dtype_state = np.dtype([
        ('proj', np.float64),
        ('state', object),
        ('hs', bool),
        ('hps', bool),
        ('pure', bool)
    ])

    def __init__(self, size):

        self._data = np.zeros(size, dtype=self._dtype_state)
        self._data['state'] = None

        self.__up_to_date = None

    def __getitem__(self, item):

        if isinstance(item, tuple):
            if Ellipsis in item:
                # self.__in_need_of_checking[item] = True
                return self.state.__getitem__(item)

            try:
                key = item[0]
                within = item[1:]
                # self.__in_need_of_checking[key] = True
                return self.state.__getitem__(key)[within]

            except IndexError:
                pass

        # self.__in_need_of_checking[item] = True
        return self.state.__getitem__(item)

    def __setitem__(self, key, value):
        # Only allowed values are density matrices, vectors, or None
        self._up_to_date = False
        self._data['hps'][key] = False

        if value is None:
            self.state[key] = None
            self.has_state[key] = False
            self.pure[key] = False

            return

        if isinstance(key, tuple):
            within = key[1:]

            if within:
                self.state[key[0]][within] = value
                self.has_state[key[0]] = True

                n = len(np.asarray(key[0]).shape)

                if n:
                    self.pure[key[0]] = self.state[key[0]][0].shape < 2
                else:
                    self.pure[key[0]] = self.state[key[0]].shape < 2

                return

            else:
                try:
                    key = key[0]
                except IndexError:
                    pass

        if isinstance(key, int):

            self.state[key] = value
            self.has_state[key] = np.bool_(value).any()
            self.pure[key] = len(self.state[key].shape) < 2

        else:

            if isinstance(value, BathState):
                self.state[key] = [rho for rho in value]
                self.has_state[key] = value.has_state
                self.pure[key] = value.pure

                return

            if not isinstance(value, np.ndarray):
                try:
                    value = np.asarray(value, dtype=np.complex128)

                except (ValueError, TypeError):
                    value = np.asarray(value, dtype=np.object_)

            if not value.dtype == np.object_:

                inshape = np.asarray(self.state[key]).shape

                if not inshape:

                    self.state[key][...].reshape(-1)[0] = value
                    self.has_state[key] = True
                    self.pure = len(value.shape) < 2

                else:
                    if len(value.shape) == 2:

                        v1 = value.shape[0]
                        v2 = value.shape[1]
                        if v1 == v2:
                            # Matrix
                            if v1 == inshape[0]:
                                if np.isclose(np.linalg.norm(value, axis=1), 1).all():
                                    # Array of vectors
                                    fshape = value.shape

                                elif np.isclose(np.trace(value), 1):
                                    # Density matrix b/c trace is 1
                                    # warnings.warn("Cannot decide whether a stack of vectors or density matrix. "
                                    #               "Assuming density matrix", stacklevel=2)
                                    fshape = inshape + value.shape
                                else:
                                    raise ValueError('Something is really wrong with this world.')
                                    # Else array of vectors

                            else:
                                # Still matrix
                                fshape = inshape + value.shape

                            self.pure[key] = False
                        else:
                            # Array of vectors
                            self.pure[key] = True
                            fshape = value.shape

                    elif len(value.shape) == 1:
                        # Vector
                        self.pure[key] = True
                        fshape = inshape + value.shape
                    else:
                        # Matrix
                        fshape = value.shape
                        self.pure[key] = False

                    broadcasted = np.array(np.broadcast_to(value, fshape))
                    self.state[key] = objarr(broadcasted)
                    self.has_state[key] = True  # [rho is not None for rho in broadcasted] Cannot be None

            else:
                self.state[key] = value
                self.has_state[key] = [rho is not None for rho in value]
                self.pure[key] = [(len(rho.shape) == 1) if rho is not None else False for rho in value]

    def gen_pure(self, rho, dim):
        """
        Generate pure states from the :math:`S_z` projections to be stored in the given ``BathState`` object.

        Args:
            rho (ndarray with shape (n, )): Array of the desired projections.
            dim (ndarray with shape (n,) ): Array of the dimensions of the spins.

        Returns:
            BathState: View of the ``BathState`` object.
        """
        rho = np.asarray(rho)

        if not rho.shape:
            # assume s is int or float showing the spin projection in the pure state
            d = dim
            if not self.shape:
                rho = vector_from_s(rho, d)
            else:
                if (d == d[0]).all():
                    d = d[0]
                    rho = np.broadcast_to(vector_from_s(rho, d), self.shape + (d,))
                else:
                    rho = [vector_from_s(rho, d_i) for d_i in d]
        else:
            rho = gen_state_list(rho, np.broadcast_to(dim, self.shape))
        self[...] = rho

        return self

    @property
    def state(self):
        """
        ndarray: Return an underlying object array.
        """
        return self._data['state']

    @property
    def pure(self):
        """
        ndarray: Bool property. True if given entry is a pure state, False otherwise.
        """
        return self._data['pure']

    @pure.setter
    def pure(self, pure):

        self._data['pure'] = pure

    @property
    def _up_to_date(self):
        """
        bool: True if projections are up to date, False otherwise.
        """
        if self.__up_to_date is None:
            self.__up_to_date = np.asarray(False)

        return self.__up_to_date

    @_up_to_date.setter
    def _up_to_date(self, item):

        if self.__up_to_date is None:
            self.__up_to_date = np.asarray(False)

        self.__up_to_date[...] = item

    @property
    def proj(self):
        """
        ndarray: Projections of bath states on :math:`S_z`.
        """
        if not self._up_to_date:
            self._project()

        return self._data['proj']

    @property
    def has_state(self):
        """
        ndarray: Bool property. True if given element was initialized as a state, False otherwise.
        """
        # if self._in_need_of_checking.any():
        #     self._has_state[self._in_need_of_checking] = False
        return self._data['hs']

    def project(self, rotation=None):
        """
        Generate projections of bath states on :math:`S_z`.

        Args:
            rotation (optional, ndarray with shape (3, 3)):
                Matrix used to transform :math:`S_z` matrix as :math:`S_z' = R^{\dagger} S_z R`.

        Returns:
            ndarray with shape (n, ): Array with projections of the state.
        """
        if rotation is None:
            self._data['hps'] = False
            self._project()
            return self._data['proj'].copy()

        projections = np.zeros(self.shape)

        if self.has_state.any():
            projections[self.has_state] = project_bath_states(self.state[self.has_state],
                                                              (not bool(self.state[self.has_state].shape)),
                                                              rotation)
        return projections

    def _project(self):

        which = self.has_state & (~self._data['hps'])
        if which.any():
            projected = project_bath_states(self.state[which], not bool(self.state[which].shape))
            self._data['proj'][which] = projected
            self._data['hps'][which] = True

        self._up_to_date = True

    def _get_state(self, item):

        obj = self.__new__(self.__class__)
        obj._data = self._data[item]
        obj.__up_to_date = self._up_to_date

        return obj

    @property
    def shape(self):
        """
        tuple: Shape of the BathState underlying array.
        """
        return self._data.shape

    @property
    def size(self):
        """
        int: Size of the BathState underlying array.
        """
        return self._data.size

    def any(self, *args, **kawrgs):
        """
        Returns the output of ``.has_state.any`` method.
        Args:
            *args: Positional arguments of ``.has_state.any`` method.
            **kawrgs: Keyword arguments of ``.has_state.any`` method.

        Returns:
            bool: True if any entry is initialized. Otherwise False.
        """
        return self.has_state.any(*args, **kawrgs)

    def __repr__(self):
        return f"{self.__class__.__name__}(" + self.state.__str__() + ")"

    def __str__(self):
        return self.state.__str__()


def objarr(array):
    """
    Make an array with object entries from iterable.

    Args:
        array (iterable): Iterable containing elements of the future array.

    Returns:
        ndarray: Object array.
    """
    nar = len(array)
    obj = np.empty(nar, dtype=object)

    for i in range(nar):
        obj[i] = np.ascontiguousarray(array[i])
    return obj


def project_bath_states(states, single=False, rotation=None):
    r"""
    Generate projections of bath states on :math:`S_z` axis from any type of states input.

    Args:
        states (array-like): Array of bath spin states.
        single (bool): True if a single bath spin. Default False.
        rotation (optional, ndarray with shape (3, 3)):
            Matrix used to transform :math:`S_z` matrix as :math:`S_z' = R^{\dagger} S_z R`.

    Returns:
        ndarray: Array of :math:`S_z` projections of the bath states.
    """
    # Ask for single b/c check against shape cannot distinguish 2x2 dm and 2 vectors of 2
    # Other checks are kinda expensive

    ndstates = np.asarray(states)

    if not ndstates.shape and ndstates.dtype == object:
        ndstates = ndstates[()]
        single = True

    projected_bath_state = None

    if ndstates.dtype == object:

        try:
            ndstates = np.stack(ndstates)

        except ValueError:
            if rotation is None:
                projected_bath_state = _loop_trace(List(states))
            else:
                projected_bath_state = _loop_trace_rotate(List(states))

    if projected_bath_state is None:
        spin = (ndstates.shape[1] - 1) / 2
        sz = _smc[spin].z

        if rotation is not None:
            sz = rotation.conj().T @ sz @ rotation

        if len(ndstates.shape) == 2 + (not single):
            # projected_bath_state = np.empty((ndstates.shape[0], 3))

            # projected_bath_state[:, 0] = np.trace(np.matmul(ndstates, _smc[spin].x), axis1=-2, axis2=-1)
            # projected_bath_state[:, 1] = np.trace(np.matmul(ndstates, _smc[spin].y), axis1=-2, axis2=-1)
            projected_bath_state = np.trace(np.matmul(ndstates, sz), axis1=-2, axis2=-1).real  # [:, 2]

        else:
            # Assume vectors
            z_psi = np.einsum('ij,...j->...i', sz, ndstates)
            projected_bath_state = np.einsum('...j,...j->...', ndstates.conj(), z_psi).real
            # projected_bath_state = ndstates

    # if len(projected_bath_state.shape) > 1 and not np.any(projected_bath_state[:, :2]):
    #     projected_bath_state = projected_bath_state[:, 2]

    return projected_bath_state


@jit(cache=True, nopython=True)
def _loop_trace(states):
    proj_states = np.empty((len(states),), dtype=np.float64)  # (len(states), 3)
    dims = List()

    # sx = List()
    # sy = List()
    sz = List()

    for j, dm in enumerate(states):
        dm = dm.astype(np.complex128)
        dim = dm.shape[0]
        try:
            ind = dims.index(dim)
        except:
            # sxnew, synew, sznew = _gen_sm(dim)
            sznew = _gen_sz(dim)
            # sx.append(sxnew)
            # sy.append(synew)
            sz.append(sznew)
            dims.append(dim)

            ind = -1
        if len(dm.shape) == 2:
            # xproj = np.trace(dm @ sx[ind])
            # yproj = np.trace(dm @ sy[ind])
            zproj = np.diag(dm @ sz[ind]).sum().real
        else:
            # xproj = dm.conj() @ sx[ind] @ dm
            # yproj = dm.conj() @ sy[ind] @ dm
            zproj = (dm.conj() @ sz[ind] @ dm).real

        # proj_states[j, 0] = xproj
        # proj_states[j, 1] = yproj
        proj_states[j] = zproj  # [j, 2]

    return proj_states


@jit(cache=True, nopython=True)
def _loop_trace_rotate(states, rotation):
    proj_states = np.empty((len(states),), dtype=np.float64)  # (len(states), 3)
    dims = List()
    sz = List()

    for j, dm in enumerate(states):
        dm = dm.astype(np.complex128)
        dim = dm.shape[0]
        try:
            ind = dims.index(dim)

        except:

            sznew = rotation.conj().T @ _gen_sz(dim) @ rotation
            sz.append(sznew)
            dims.append(dim)
            ind = -1

        if len(dm.shape) == 2:

            dm = dm
            zproj = np.diag(dm @ sz[ind]).sum().real

        else:
            zproj = (dm.conj() @ sz[ind] @ dm).real

        proj_states[j] = zproj

    return proj_states


@jit(nopython=True)
def _gen_sz(dim):
    s = (dim - 1) / 2
    projections = np.linspace(-s, s, dim).astype(np.complex128)
    return np.diag(projections[::-1])



import numpy as np
from numpy import ma as ma
from pycce.bath.array import BathArray
from pycce.constants import PI2
from pycce.h import total_hamiltonian
from pycce.run import simple_propagator
from pycce.run.base import RunObject, generate_initial_state
from pycce.utilities import shorten_dimensions, outer


def rotation_propagator(u, rotations):
    """
    Generate the propagator from the simple propagator and set of :math:`2\tau` equispaced rotation operators.

    .. note::

        While the spacing between rotation operators is assumed to be :math:`2\tau`, the spacing before and after
        the first and the last rotation respectively is assumed to be :math:`\tau`.

    Args:
        u (ndarray with shape (n, N, N)): Simple propagator.
        rotations (ndarray with shape (x, N, N)): Array of rotation operators.

    Returns:
        ndarray with shape (n, N, N): Full propagator.

    """
    full_u = np.eye(u.shape[1], dtype=np.complex128)

    for rotation in rotations:
        full_u = np.matmul(u, full_u)

        if rotation is not None:
            full_u = np.matmul(rotation, full_u)

        full_u = np.matmul(u, full_u)

    return full_u


def _get_state(state, center):
    if callable(state):
        return state

    state = np.asarray(state)
    if state.size == 1:
        return center.eigenvectors[:, int(state)]
    else:
        return state.astype(np.complex128)


class gCCE(RunObject):
    """
    Class for running generalized CCE simulations.

    .. note::

        Subclass of the ``RunObject`` abstract class.

    Args:
        *args: Positional arguments of the ``RunObject``.

        pulses (Sequence): Sequence object, containing series of pulses, applied to the system.

        as_delay (bool):
            True if time points are delay between pulses, False if time points are total time.

        fulldm (bool):
            True if return full density matrix. Default False.

        **kwargs: Keyword arguments of the ``RunObject``.

    """

    def __init__(self, *args, i=None, j=None, as_delay=False, fulldm=False, normalized=True, **kwargs):

        self.as_delay = as_delay
        """ bool: True if time points are delay between pulses, False if time points are total time."""

        self.dm0 = None
        """ ndarray with shape (2s+1, 2s+1): Initial density matrix of the central spin."""
        self.normalization = None
        """ float: Coherence at time 0."""
        self.zero_cluster = None
        """ ndarray with shape (n,): Coherence computed for the isolated central spin."""
        self.i = i
        self.j = j
        self.alpha = None
        self.beta = None
        self.normalized = normalized
        self.fulldm = fulldm
        """ bool: True if return full density matrix."""

        super().__init__(*args, **kwargs)

    def preprocess(self):
        super().preprocess()

        # Emulate kernel
        self.base_hamiltonian = self.center.generate_hamiltonian(magnetic_field=self.magnetic_field)
        self.cluster = BathArray((0,))
        self.cluster_indexes = np.array([], dtype=int)

        # if self.has_states:
        #     self.others = self.bath
        #     self.others_mask = np.ones(self.bath.size, dtype=bool)

        self._check_hamiltonian()

        self.dm0 = self.center.state
        self.normalization = outer(self.center.state, self.center.state)

        if self.i is None:
            alpha = self.center.alpha
        else:
            alpha = _get_state(self.i, self.center)

        if self.j is None:
            beta = self.center.beta
        else:
            beta = _get_state(self.j, self.center)

        if self.pulses is not None:
            self.center.generate_sigma()
            self.generate_pulses()

            for rotation in self.rotations:
                if rotation is not None:
                    if self.i is None:
                        alpha = rotation @ alpha
                    if self.j is None:
                        beta = rotation @ beta

                    self.normalization = rotation @ self.normalization @ rotation.T.conj()

        self.alpha = alpha
        self.beta = beta
        self.zero_cluster = 1  # For compute result to work

        self.zero_cluster = self.compute_result()

        if self.fulldm:
            self.zero_cluster = ma.array(self.zero_cluster, mask=np.isclose(self.zero_cluster, 0), fill_value=0j)
        elif self.normalized:
            self.normalization = self.process_dm(self.normalization)

        # else:
        #     density_matrix = self.center.eigenvectors.conj().T @ density_matrix @ self.center.eigenvectors

    def process_dm(self, density_matrix):
        """
        Obtain the result from the density matrices.

        Args:
            density_matrix (ndarray with shape (n, N, N)): Array of the density matrices.

        Returns:
            ndarray:
                Depending on the parameters,
                returns the off diagonal element of the density matrix or full matrix.
        """
        if self.fulldm:
            return density_matrix

        if callable(self.alpha):
            result = self.alpha(density_matrix)
        elif callable(self.beta):
            result = self.beta(density_matrix)
        else:
            result = self.alpha.conj() @ density_matrix @ self.beta

        return result

    def postprocess(self):
        self.result = self.zero_cluster * self.result
        if self.fulldm:
            self.result = self.result.filled()
        elif self.normalized:
            self.result = self.result / self.normalization

        super().postprocess()

        # else:
        #     self.result = self.center.eigenvectors @ self.result @ self.center.eigenvectors.conj().T

    def generate_hamiltonian(self):
        """
        Using the attributes of the ``self`` object,
        compute the cluster hamiltonian including the central spin.

        Returns:
            Hamiltonian: Cluster hamiltonian.

        """
        ham = total_hamiltonian(self.cluster, self.center, self.magnetic_field)

        return ham

    def compute_result(self):
        """
        Using the attributes of the ``self`` object,
        compute the coherence function of the central spin.

        Returns:

            ndarray: Computed coherence.

        """

        dimensions = shorten_dimensions(self.base_hamiltonian.dimensions, self.center.size)

        initial_state = generate_initial_state(dimensions, states=self.states, central_state=self.dm0)

        unitary_evolution = self.propagator()

        if initial_state.ndim > 1:
            # rho U^\dagger
            dm_udagger = np.matmul(initial_state, unitary_evolution.conj().transpose(0, 2, 1))
            # U rho U^\dagger
            result = np.matmul(unitary_evolution, dm_udagger)
        else:
            # |dm> = U|dm>
            result = unitary_evolution @ initial_state
            # |dm><dm|
            result = np.einsum('ki,kj->kij', result, result.conj())

        initial_shape = result.shape
        result.shape = (initial_shape[0], *dimensions, *dimensions)

        for d in range(len(dimensions) + 1, 2, -1):  # The last one is el spin
            result = np.trace(result, axis1=1, axis2=d)

            if result.shape[1:] == self.dm0.shape:  # break if shape is the same
                break

        result = self.process_dm(result)

        return result / self.zero_cluster

    def propagator(self):
        """
        Function to compute time propagator U.

        Returns:
            ndarray with shape (t, n, n): Array of propagators, evaluated at each time point in ``self.timespace``.
        """
        if not self.pulses:
            return simple_propagator(self.timespace, self.hamiltonian)

        if self.delays is None:
            if self.projected_states is None:
                return self._no_delays_no_ps()
            # proj_states is not None - there are bath rotations alas
            return self._no_delays_ps()

        # There are delays but no bath flips
        if self.projected_states is None:
            return self._delays_no_ps()

        # The most complicated case - both projected_states is not None and delays is not None
        return self._delays_ps()

    def _no_delays_no_ps(self):

        delays = self.timespace if self.as_delay else self.timespace / (2 * len(self.pulses))

        # Same propagator for all parts
        u = simple_propagator(delays, self.hamiltonian)

        return rotation_propagator(u, self.rotations)

    def _no_delays_ps(self):
        delays = self.timespace if self.as_delay else self.timespace / (2 * len(self.pulses))

        self.get_hamiltonian_variable_bath_state(0)
        u = simple_propagator(delays, self.hamiltonian)

        full_u = np.eye(self.base_hamiltonian.data.shape[0], dtype=np.complex128)

        ps_counter = 0

        for p, rotation in zip(self.pulses, self.rotations):

            full_u = np.matmul(u, full_u)

            if rotation is not None:
                full_u = np.matmul(rotation, full_u)

                if p.bath_names is not None:
                    ps_counter += 1
                    self.get_hamiltonian_variable_bath_state(ps_counter)
                    u = simple_propagator(delays, self.hamiltonian)

            full_u = np.matmul(u, full_u)

        return full_u

    def _delays_no_ps(self):

        evalues, evec = np.linalg.eigh(self.hamiltonian * PI2)

        full_u = np.eye(self.base_hamiltonian.data.shape[0], dtype=np.complex128)

        for delay, rotation in zip(self.delays, self.rotations):
            eigexp = np.exp(-1j * np.outer(delay, evalues),
                            dtype=np.complex128)

            u = np.matmul(np.einsum('ij,kj->kij', evec, eigexp, dtype=np.complex128),
                          evec.conj().T)

            full_u = np.matmul(u, full_u)

            if rotation is not None:
                full_u = np.matmul(rotation, full_u)

            full_u = np.matmul(u, full_u)

        return full_u

    def _delays_ps(self):
        self.get_hamiltonian_variable_bath_state(0)

        full_u = np.eye(self.hamiltonian.shape[0], dtype=np.complex128)

        ps_counter = 0
        times = 0

        for p, rotation, delay in zip(self.pulses, self.rotations, self.delays):
            times += delay
            u = simple_propagator(delay, self.hamiltonian)
            full_u = np.matmul(u, full_u)

            if rotation is not None:
                full_u = np.matmul(rotation, full_u)

                if p.bath_names is not None:
                    ps_counter += 1
                    self.get_hamiltonian_variable_bath_state(ps_counter)

        which = np.isclose(self.timespace, times)

        if ((self.timespace - times)[~which] >= 0).all():
            u = simple_propagator(self.timespace - times, self.hamiltonian)

            full_u = np.matmul(u, full_u)

        elif not which.all():
            raise ValueError(f"Pulse sequence time steps add up to larger than total times. Delays at"
                             f"{self.timespace[(self.timespace - times) < 0]} ms are longer than total time.")

        return full_u

# def propagator(timespace, hamiltonian,
#                pulses=None, as_delay=False):

#     evalues, evec = np.linalg.eigh(hamiltonian * PI2)
#
#     if not pulses:
#
#         eigexp = np.exp(-1j * np.outer(timespace, evalues),
#                         dtype=np.complex128)
#
#         u = np.matmul(np.einsum('ij,kj->kij', evec, eigexp, dtype=np.complex128),
#                       evec.conj().T)
#
#         return u
#
#     else:
#
#         if pulses.delays is None:
#             if not as_delay:
#                 n = len(pulses)
#                 timespace = timespace / (2 * n)
#
#             eigexp = np.exp(-1j * np.outer(timespace, evalues),
#                             dtype=np.complex128)
#
#             u = np.matmul(np.einsum('ij,kj->kij', evec, eigexp, dtype=np.complex128),
#                           evec.conj().T)
#             U = np.eye(u.shape[1], dtype=np.complex128)
#
#             for rotation in pulses.rotations:
#                 U = np.matmul(u, U)
#                 if rotation is not None:
#                     U = np.matmul(rotation, U)
#                 U = np.matmul(u, U)
#
#             return U
#
#         U = None
#
#         times = 0
#         for timesteps, rotation in zip(pulses.delays, pulses.rotations):
#
#             eigexp = np.exp(-1j * np.outer(timesteps, evalues),
#                             dtype=np.complex128)
#
#             u = np.matmul(np.einsum('...ij,...j->...ij', evec, eigexp, dtype=np.complex128),
#                           evec.conj().T)
#             times += timesteps
#
#             if U is None:
#                 if rotation is not None:
#                     U = np.matmul(rotation, u)
#                 else:
#                     U = u
#
#             else:
#                 U = np.matmul(u, U)
#                 if rotation is not None:
#                     U = np.matmul(rotation, U)
#
#         which = np.isclose(timespace, times)
#
#         if ((timespace - times)[~which] >= 0).all():
#             eigexp = np.exp(-1j * np.outer(timespace - times, evalues),
#                             dtype=np.complex128)
#
#             u = np.matmul(np.einsum('ij,kj->kij', evec, eigexp, dtype=np.complex128),
#                           evec.conj().T)
#
#             U = np.matmul(u, U)
#         elif not which.all():
#             raise ValueError(f"Pulse sequence time steps add up to larger than total times. Delays at"
#                              f"{timespace[(timespace - times) < 0]} ms are longer than total time.")
#     return U
#
#
# def jit_propagator_advanced(timespace, hamiltonian,
#                             total_delays, total_rotations):
#     evalues, evec = np.linalg.eigh(hamiltonian * PI2)
#
#     times = np.zeros(timespace.shape)
#     propagators = np.zeros(timespace.shape + hamiltonian.shape)
#     eye = np.eye(hamiltonian.shape[0], dtype=np.complex128)
#
#     for index in range(timespace.size):
#         u = eye
#
#         total_time = timespace[index]
#         delays = total_delays[:, index]
#
#         applicable_rotations = delays <= total_time
#         delays = delays[applicable_rotations]
#         rotations = total_rotations[applicable_rotations]
#         order = np.argsort(delays)
#         passed_time = 0
#
#         for i in order:
#             rotation = rotations[i]
#             t = delays[i] - passed_time
#             if t:
#                 eigexp = np.exp(-1j * t * evalues, dtype=np.complex128)
#                 u = (evec @ np.diag(eigexp) @ evec.conj().T) @ u
#                 passed_time = delays[i]
#
#             if not (rotation == eye).all():
#                 u = rotation @ u
#
#         propagators[index] = u
#
#     return propagators

#
# def compute_dm(initial_state, H, timespace, pulse_sequence=None, as_delay=False, states=None, ncenters=1):
#     """
#     Function to compute density matrix of the central spin, given Hamiltonian H.
#
#     Args:
#         initial_state (ndarray): Initial density matrix of central spin.
#
#         H (ndarray): Cluster Hamiltonian.
#
#         timespace (ndarray): Time points at which to compute density matrix.
#
#         pulse_sequence (Sequence): Sequence of pulses.
#
#         as_delay (bool):
#             True if time points are delay between pulses, False if time points are total time.
#
#         states (ndarray): ndarray of bath states in any accepted format.
#
#         ncenters (int): Number of central spins.
#
#     Returns:
#         ndarray: Array of density matrices evaluated at all time points in timespace.
#     """
#     center_shape = initial_state.shape
#     dimensions = shorten_dimensions(H.dimensions, ncenters)
#
#     initial_state = generate_initial_state(dimensions, states=states, central_state=initial_state)
#     dm = full_dm(initial_state, H, timespace, pulse_sequence=pulse_sequence, as_delay=as_delay)
#     initial_shape = dm.shape
#     dm.shape = (initial_shape[0], *dimensions, *dimensions)
#     for d in range(len(dimensions) + 1, 2, -1):  # The last one is el spin
#         dm = np.trace(dm, axis1=1, axis2=d)
#         if dm.shape[1:] == center_shape:  # break if shape is the same
#             break
#     return dm
#
#
# def full_dm(dm0, H, timespace, pulse_sequence=None, as_delay=False):
#     """
#     A function to compute density matrix of the cluster, using hamiltonian H
#     from the initial density matrix of the cluster.
#
#     Args:
#         dm0 (ndarray):
#             Initial density matrix of the cluster
#         H (ndarray):
#             Cluster Hamiltonian
#         timespace (ndarray): Time points at which to compute coherence function.
#         pulse_sequence (Sequence): Sequence of pulses.
#
#         as_delay (bool):
#             True if time points are delay between pulses, False if time points are total time. Ignored if delays
#             are provided in the ``pulse_sequence``.
#
#     Returns:
#         ndarray: Array of density matrices of the cluster, evaluated at the time points from timespace.
#     """
#     dm = 0
#     return dm



import copy
from collections import defaultdict
from collections.abc import MutableMapping

import numpy as np
from scipy.sparse.sputils import isintlike

two_layer_dict = lambda: defaultdict(dict)


class InteractionMap(MutableMapping):
    """
    Dict-like object containing information about tensor interactions between two spins.

    Each key is a tuple of two spin indexes.

    Args:
        rows (array-like with shape (n, )):
            Indexes of the bath spins, appearing on the left in the pairwise interaction.
        columns (array-like with shape (n, )):
            Indexes of the bath spins, appearing on the right in the pairwise interaction.
        tensors (array-like with shape (n, 3, 3)):
            Tensors of pairwise interactions between two spins with the indexes in ``rows`` and ``columns``.
    Attributes:
        mapping (dict): Actual dictionary storing the data.
    """

    def __init__(self, rows=None, columns=None, tensors=None):
        self.mapping = dict()
        self._indexes = None
        self._data = None
        if (rows is not None) & (columns is not None) & (tensors is not None):
            self[rows, columns] = tensors
            self.__gen_indexes()

    @property
    def indexes(self):
        """
        ndarray with shape (n, 2): Array with the indexes of pairs of spins, for which the tensors are stored.
        """
        if self._indexes is None:
            self.__gen_indexes()
        return self._indexes

    @indexes.setter
    def indexes(self, newindexes):
        self._indexes = newindexes

    @property
    def data(self):
        if self._data is None:
            self.__gen_data()
        return self._data

    def __getitem__(self, key):
        a, b = _index(key)
        try:
            if a < b:
                return self.mapping[a, b]
            else:
                return self.mapping[b, a].T
        except (TypeError, ValueError):
            try:
                vs = []
                for j, k in zip(a, b):
                    v = self[j, k]
                    vs.append(v)
                return np.asarray(vs)

            except TypeError as e:
                raise TypeError('invalid index format') from e

    def __delitem__(self, key):
        a, b = _index(key)

        if isintlike(a) and isintlike(b):
            if a < b:
                del self.mapping[a, b]
            else:
                del self.mapping[b, a]
        else:
            for j, k in zip(a, b):
                if j < k:
                    del self.mapping[j, k]
                else:
                    del self.mapping[k, j]

        self.indexes = None
        self._data = None

    def __setitem__(self, key, value):
        value = np.asarray(value, dtype=np.float64)

        if value.size == 9:
            value = value.reshape(3, 3)
        elif value.size // 9 > 0:
            value = value.reshape(-1, 3, 3)
        a, b = _index(key)

        if isintlike(a) and isintlike(b):
            assert value.size == 9, 'Tensor should have shape of (3, 3)'
            if a < b:
                self.mapping[a, b] = value
            else:
                self.mapping[b, a] = value.T

        else:
            try:
                if value.size == 9:
                    for j, k in zip(a, b):
                        if j < k:
                            self.mapping[j, k] = value
                        else:
                            self.mapping[k, j] = value.T
                else:
                    for j, k, v in zip(a, b, value):
                        if j < k:
                            self.mapping[j, k] = v
                        else:
                            self.mapping[k, j] = v.T

            except TypeError as e:
                raise TypeError('invalid index format') from e
        self._data = None
        self.indexes = None

    def shift(self, start, inplace=True):
        """
        Add an offset ``start`` to the indexes. If ``inplace`` is False, returns the copy of InteractionMap.

        Args:
            start (int): Offset in indexes.
            inplace (bool): If True, makes changes inplace. Otherwise returns copy of the map.

        Returns:
            InteractionMap: Map with shifted indexes.
        """

        if inplace:
            imap = self
        else:
            imap = copy.deepcopy(self)

        for i, j in imap.indexes:
            imap.mapping[i + start, j + start] = imap.mapping.pop((i, j))

        imap.indexes = None
        imap._data = None
        return imap

    def __iter__(self):
        return iter(self.mapping)

    def __len__(self):
        return len(self.indexes)

    def __repr__(self):
        return f"{type(self).__name__}({self.mapping})"

    # def update(self, **kwargs):
    #     # Prevent direct usage of update
    #     raise NotImplementedError("Direct modification to InteractionMap element "
    #                               "is not allowed.")

    def keys(self):
        return self.mapping.keys()

    def items(self):
        return self.mapping.items()

    def __gen_indexes(self):
        self.indexes = np.fromiter((ind for pair in self.mapping.keys() for ind in pair),
                                   dtype=np.int32).reshape(-1, 2)

    def __gen_data(self):
        length = self.indexes.shape[0]

        self._data = np.zeros((length, 3, 3), dtype=np.float64)

        for j, pair in enumerate(self.indexes):
            self._data[j] = self[pair]

    def subspace(self, array):
        r"""
        Get new InteractionMap with indexes readressed from array. Within the subspace indexes are renumbered.

        Examples:

            The subspace of [3,4,7] indexes will contain InteractionMap only within [3,4,7] elements
            with new indexes [0, 1, 2].

                >>> import numpy as np
                >>> im = InteractionMap()
                >>> im[0, 3] = np.eye(3)
                >>> im[3, 7] = np.ones(3)
                >>> for k in im: print(k, '\n', im[k],)
                (0, 3)
                [[1. 0. 0.]
                 [0. 1. 0.]
                 [0. 0. 1.]]
                (3, 7)
                 [[1. 1. 1.]
                  [1. 1. 1.]
                  [1. 1. 1.]]
                >>> array = [3, 4, 7]
                >>> sim = im.subspace(array)
                >>> for k in sim: print(k, '\n', sim[k])
                (0, 2)
                [[1. 1. 1.]
                 [1. 1. 1.]
                 [1. 1. 1.]]


        Args:
            array (ndarray): Either bool array containing True for elements within the subspace
                or array of indexes presented in the subspace.

        Returns:
            InteractionMap: The map for the subspace.
        """

        array = np.asarray(array)

        if array.dtype == bool:
            ind = np.arange(array.size, dtype=np.int32)
            indexes = ind[array]
        else:
            indexes = array

        newindexes = np.arange(indexes.size, dtype=np.int32)

        where = np.isin(self.indexes[:, 0], indexes) & np.isin(self.indexes[:, 1], indexes)
        pairs = self.indexes[where]

        xind = indexes.argsort()
        ypos = np.searchsorted(indexes[xind], pairs.flatten())

        indices = xind[ypos]
        newpairs = newindexes[indices].reshape(-1, 2)

        newdict = {}

        for (oi, oj), (i, j) in zip(pairs, newpairs):
            if i < j:
                newdict[i, j] = self[oi, oj]
            else:
                newdict[j, i] = self[oj, oi]

        return InteractionMap.from_dict(newdict, presorted=True)

    @classmethod
    def from_dict(cls, dictionary, presorted=False):
        """
        Generate InteractionMap from the dictionary.

        Args:

            dictionary (dict): Dictionary with tensors.

            presorted (bool): If true, assumes that the keys in the dictionary were already presorted.

        Returns:

            InteractionMap: New instance generated from the dictionary.
        """
        obj = cls()
        if presorted:
            obj.mapping = dictionary
        else:
            for k in dictionary:
                obj[k] = dictionary[k]
        return obj

    def __add__(self, other):
        new_obj = InteractionMap()
        keys_1 = list(self.keys())
        keys_2 = list(other.keys())

        for k in {*keys_1, *keys_2}:

            if (k in keys_1) and (k in keys_2):
                assert (other[k] == self[k]).all(), f'Error, tensor {k} has different properties in provided mappings'
                new_obj[k] = self[k]

            elif k in keys_1:
                new_obj[k] = self[k]
            else:
                new_obj[k] = other[k]
        return new_obj

    # TODO implement compressed tensors
    def _compress(self):
        self.tensors = np.empty((self.indexes.shape[0], 3, 3), dtype=np.float64)

        for j, ind in enumerate(self.indexes):
            self.tensors[j] = self[ind]


# This implementation is slower than the one above :(
class _CompressedIMap:
    def __init__(self, indexes=None, tensors=None, imap=None):
        self.tensors = tensors
        self.indexes = indexes
        if self.indexes is not None:
            unsorted = self.indexes[:, 1] < self.indexes[:, 0]
            if unsorted.any():
                self.tensors[unsorted] = np.transpose(self.tensors[unsorted], axes=(0, 2, 1))
                self.indexes[unsorted] = self.indexes[unsorted][..., ::-1]
        if imap is not None:
            self.tensors = np.empty((imap.indexes.shape[0], 3, 3), dtype=np.float64)
            self.indexes = imap.indexes.copy()

            for j, ind in enumerate(imap.indexes):
                self.tensors[j] = imap[ind]

    def subspace(self, array):

        array = np.asarray(array)

        if array.dtype == bool:
            ind = np.arange(array.size, dtype=np.int32)
            indexes = ind[array]
        else:
            # Indexes of the subspace. E.g. 1, 3, 5, 7
            indexes = array

        # Indexes in the subspace 0, 1, 2, 3
        newindexes = np.arange(indexes.size, dtype=np.int32)
        # Which indexes are present
        full = (self.indexes[:, :, np.newaxis] == indexes[np.newaxis, np.newaxis, :])
        where = full.any(-1).all(-1)
        where_exactly = np.nonzero(full[where])
        # lastpairs[where_exactly[:-1]] = newindexes[where_exactly[-1]]
        newpairs = newindexes[where_exactly[-1]].reshape(-1, 2)
        newtensors = self.tensors[where]
        return _CompressedIMap(newpairs, newtensors)

    def __getitem__(self, key):

        if not key.shape[-1] == 2:
            raise KeyError

        unsorted = (key[..., :-1] > key[..., 1:])
        any_unsorted = unsorted.any()

        if any_unsorted:
            key[unsorted] = key[unsorted][..., ::-1]
        which = (self.indexes[0] == key).all(1)

        if not np.count_nonzero(which) == key.size // 2:
            raise KeyError

        tensors = self.tensors[which]

        if any_unsorted:
            tensors[unsorted] = np.transpose(tensors[unsorted], axes=(0, 2, 1))
        if key.size == 2:
            return tensors[0]

        return tensors

    def __setitem__(self, key, value):
        value = np.asarray(value)
        which = (self.indexes == key).all(1)
        if not which.any():
            raise KeyError
        key.argsort(1)
        value.sort()


def _index(key):
    try:
        a, b = key
    except TypeError as e:
        raise TypeError('invalid index format') from e
    return a, b



import re
import warnings

import numpy as np
from pycce.bath.array import BathArray, transform
from pycce.constants import MHZ_TO_KHZ, EFG_CONVERSION

from .base import DFTCoordinates, find_first_index, yield_index, set_isotopes


def read_orca(fname, isotopes=None, types=None, center=None,
              find_isotopes=True, rotation_matrix=None, rm_style='col'):
    r"""
    Function to read ORCA output containing the hyperfines couplings and EFG tensors.

    if ``find_isotopes`` is set to True changes the names of the atoms to the most abundant isotopes.
    If that is not the desired outcome, user can define which isotopes to use using keyword isotopes.

    Args:
        fname (str): file name of the ORCA output.
        isotopes (dict): Optional.
            Dictionary with entries::

            {"element" : "isotope"}

            where "element" is the name of the element in DFT output, "isotope" is the name of the isotope.

        types (SpinDict or list of tuples): SpinDict containing SpinTypes of isotopes or input to make one.
        center (ndarray of shape (3,)): position of (0, 0, 0) point in the DFT coordinates.
        rotation_matrix (ndarray of shape (3,3)):
            Rotation matrix to rotate basis. For details see utilities.transform.
        rm_style (str):
            Indicates how rotation matrix should be interpreted.
            Can take values "col" or "row". Default "col"
        find_isotopes (bool): If true, sets isotopes instead of names of the atoms.

    Returns:
        BathArray:
            Array of bath spins with hyperfine couplings and quadrupole tensors from Orca output.

    """
    lines = open(fname).readlines()
    output = ORCACoordinates(lines)
    with warnings.catch_warnings(record=True):
        atoms = BathArray(array=output.coordinates, names=output.names, types=types)

    start = find_first_index('ELECTRIC AND MAGNETIC HYPERFINE STRUCTURE', lines)

    efgs = []
    hyperfines = []
    a_indexes = []
    q_indexes = []
    for ind in yield_index('Nucleus', lines, start=start, case_sensitive=True):
        sline = lines[ind].split()
        atom_index = int(re.sub(r'\D', '', sline[1]))
        atom_name = re.sub(r"\d+", "", sline[1])

        # assert atoms[atom_index]['N'] == atom_name, (f"Name mismatch at {atom_index}{atom_name}"
        #                                              f" {atoms[atom_index]['N']} expected")

        try:
            n = find_first_index('Raw HFC matrix', lines, start=ind) + 2
            tensor = []
            for _ in range(3):
                line = lines[n]
                tensor.append([float(x) * MHZ_TO_KHZ for x in line.split()])
                n += 1
            hyperfines.append(tensor)
            a_indexes.append(atom_index)
        except KeyError:
            pass

        try:
            n = find_first_index('Raw EFG matrix', lines, start=ind) + 1
            tensor = []
            for _ in range(3):
                line = lines[n]
                tensor.append([float(x) * EFG_CONVERSION for x in line.split()])
                n += 1
            efgs.append(tensor)
            q_indexes.append(atom_index)
        except KeyError:
            pass

    if hyperfines:
        atoms.A[a_indexes] = hyperfines

    if find_isotopes:
        set_isotopes(atoms, isotopes=isotopes, spin_types=types)

    if efgs:
        atoms[q_indexes] = atoms[q_indexes].from_efg(efgs)

    atoms = transform(atoms, center, rotation_matrix=rotation_matrix, style=rm_style)

    return atoms


class ORCACoordinates(DFTCoordinates):
    r"""
    Coordinates of the system from the ORCA output. Subclass of the DFTCoordinates.

    With initialization reads output of the ORCA.

    Args:
        orca_output (str or list of str): either name of the output file or list of lines read from that file.
    Attributes:
        alat (float): The lattice parameter in angstrom.
        cell (ndarray with shape (3, 3)):
            cell is 3x3 matrix with entries:

            .. math::

                [&[a_x\ b_x\ c_x]\\
                &[a_y\ b_y\ c_y]\\
                &[a_z\ b_z\ c_z]]

            where a, b, c are crystallographic vectors,
            and x, y, z are their coordinates in the cartesian reference frame.
        coordinates (ndarray with shape (n, 3)): array with the coordinates of atoms in the cell.
        names (ndarray with shape (n,)): array with the names of atoms in the cell.
        cell_units (str): Units of cell coordinates: 'bohr', 'angstrom', 'alat'.
        coordinates_units (str): Units of atom coordinates: 'crystal', 'bohr', 'angstrom', 'alat'.
    """

    def __init__(self, orca_output):
        super().__init__()
        self.cell = np.eye(3)
        self.cell_units = 'angstrom'
        self.read_output(orca_output)

    def read_output(self, orca_output):
        """
        Method to read coordinates of atoms from ORCA output into the ORCACoordinates instance.

        Args:
            orca_output (str or list of str): either name of the output file or list of lines read from that file.
        """
        try:
            lines = open(orca_output).readlines()

        except TypeError:
            lines = orca_output

        index = find_first_index('CARTESIAN COORDINATES (ANGSTROEM)', lines) + 1
        names = []
        coordinates = []

        while True:
            try:
                index += 1

                row_split = lines[index].split()
                name = row_split[0]
                crow = [float(x) for x in row_split[1:]]

                names.append(name)
                coordinates.append(crow)
            except (IndexError, ValueError):
                break

        self.coordinates_units = 'angstrom'
        self.coordinates = np.asarray(coordinates)
        self.names = np.array(names, dtype='<U16')
